import os
from tests.utils_for_test import do_log_tests


def test_arcane_blast_v45():
    json_fname = "arcane_blast_v45.json"
    json_path = os.path.join(os.path.dirname(__file__), "jsons", json_fname)
    do_log_tests(json_path)


def test_lumenstone_enternalflow_v45():
    json_fname = "lumenstone_enternalflow_v45.json"
    json_path = os.path.join(os.path.dirname(__file__), "jsons", json_fname)
    do_log_tests(json_path)


def test_balance_v45():
    json_fname = "balance_v45.json"
    json_path = os.path.join(os.path.dirname(__file__), "jsons", json_fname)
    do_log_tests(json_path)


def test_meropide_golden_v45():
    json_fname = "meropide_golden_v45.json"
    json_path = os.path.join(os.path.dirname(__file__), "jsons", json_fname)
    do_log_tests(json_path)


def test_coverage_improve_v45():
    json_fname = "coverage_improve_v45.json"
    json_path = os.path.join(os.path.dirname(__file__), "jsons", json_fname)
    do_log_tests(json_path)


def test_neuvillette_v45():
    json_fname = "neuvillette_v45.json"
    json_path = os.path.join(os.path.dirname(__file__), "jsons", json_fname)
    do_log_tests(json_path)


def test_kirara_v45():
    json_fname = "kirara_v45.json"
    json_path = os.path.join(os.path.dirname(__file__), "jsons", json_fname)
    do_log_tests(json_path)


def test_charlotte_v45():
    json_fname = "charlotte_v45.json"
    json_path = os.path.join(os.path.dirname(__file__), "jsons", json_fname)
    do_log_tests(json_path)


def test_electro_cicin_mage_v45():
    json_fname = "electro_cicin_mage_v45.json"
    json_path = os.path.join(os.path.dirname(__file__), "jsons", json_fname)
    do_log_tests(json_path)


def test_electro_cicin_mage_2_v45():
    json_fname = "electro_cicin_mage_2_v45.json"
    json_path = os.path.join(os.path.dirname(__file__), "jsons", json_fname)
    do_log_tests(json_path)


if __name__ == "__main__":
    test_arcane_blast_v45()
    test_balance_v45()
    test_lumenstone_enternalflow_v45()
    test_meropide_golden_v45()
    test_coverage_improve_v45()
    test_neuvillette_v45()
    test_kirara_v45()
    test_charlotte_v45()
    test_electro_cicin_mage_v45()



from typing import Any, Literal, List

from ....utils.class_registry import register_base_class

from ...struct import Cost

from ...modifiable_values import (
    DamageDecreaseValue,
    DamageElementEnhanceValue,
    DamageValue,
)
from ...consts import (
    ELEMENT_TO_ENCHANT_ICON,
    DamageElementalType,
    DamageType,
    ElementType,
    IconType,
    ObjectType,
    PlayerActionLabels,
)
from ..base import StatusBase
from ...action import (
    MakeDamageAction,
    RemoveObjectAction,
    Actions,
    SkipPlayerActionAction,
    UseSkillAction,
)
from ...event import (
    ChooseCharacterEventArguments,
    MakeDamageEventArguments,
    PlayerActionStartEventArguments,
    RoundEndEventArguments,
    RoundPrepareEventArguments,
    SwitchCharacterEventArguments,
    UseSkillEventArguments,
)


class CharacterStatusBase(StatusBase):
    """
    Base class of character status.
    """

    type: Literal[ObjectType.CHARACTER_STATUS] = ObjectType.CHARACTER_STATUS
    name: str
    version: str
    usage: int
    max_usage: int


register_base_class(CharacterStatusBase)


class UsageCharacterStatus(CharacterStatusBase):
    """
    Base class of team status that based on usages.
    It will implement a method to check if run out of usage; when run out
    of usage, it will remove itself.

    By default, it listens to MAKE_DAMAGE event to check if run out of usage,
    as most usage-based status will decrease usage when damage made.
    Not using AFTER_MAKE_DAMAGE because we should remove the status before
    side effects of elemental reaction, which is triggered on RECEIVE_DAMAGE.

    If it is not a damage related status, check_should_remove should be
    called manually.
    """

    name: str
    version: str
    usage: int
    max_usage: int

    def check_should_remove(self) -> List[RemoveObjectAction]:
        """
        Check if the status should be removed.
        when usage has changed, call this function to check if the status
        should be removed.
        """
        if self.usage <= 0:
            return [
                RemoveObjectAction(
                    object_position=self.position,
                )
            ]
        return []

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        When damage made, check whether the team status should be removed.
        """
        return self.check_should_remove()


class RoundCharacterStatus(CharacterStatusBase):
    """
    Base class of team status that based on rounds.  It will implement the
    event trigger on ROUND_PREPARE to check if run out of usages; when run out
    of usages, it will remove itself.
    """

    name: str
    version: str
    usage: int
    max_usage: int

    def check_should_remove(self) -> List[RemoveObjectAction]:
        """
        Check if the status should be removed.
        when round has changed, call this function to check if the status
        should be removed.
        """
        if self.usage <= 0:
            return [
                RemoveObjectAction(
                    object_position=self.position,
                )
            ]
        return []

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        When reaching prepare, decrease usage and check whether the team
        status should be removed.
        """
        self.usage -= 1
        return list(self.check_should_remove())


class DefendCharacterStatus(UsageCharacterStatus):
    """
    Base class of defend status (purple shield), decrease damage with its rule
    when receive damage, and decrease usage by 1.
    """

    name: str
    version: str
    usage: int
    max_usage: int
    min_damage_to_trigger: int
    max_in_one_time: int
    decrease_usage_by_damage: bool = False

    icon_type: Literal[IconType.BARRIER] = IconType.BARRIER

    def value_modifier_DAMAGE_DECREASE(
        self, value: DamageDecreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageDecreaseValue:
        """
        Decrease damage with its rule, and decrease 1 usage.
        """
        if not value.is_corresponding_character_receive_damage(
            self.position,
            match,
        ):
            # not this character receive damage, not modify
            return value
        if self.usage <= 0:
            return value
        new_usage = value.apply_shield(
            self.usage,
            self.min_damage_to_trigger,
            self.max_in_one_time,
            self.decrease_usage_by_damage,
        )
        assert mode == "REAL"
        self.usage = new_usage
        return value


class ShieldCharacterStatus(DefendCharacterStatus):
    """
    Base class of shield status (yellow shield), decrease damage by its usage
    and decrease corresponding usage.
    """

    name: str
    version: str
    usage: int
    max_usage: int

    icon_type: Literal[IconType.SHIELD] = IconType.SHIELD

    # papams passing to apply_shield is fixed
    min_damage_to_trigger: int = 0
    max_in_one_time: int = 0
    decrease_usage_by_damage: bool = True


class ElementalInfusionCharacterStatus(CharacterStatusBase):
    """
    Base class of doing elemental infusion. It will infuse all physical damage
    to corresponding elemental damage for this character. If there extra
    conditions, filter the condition and call value modifier in subclasses.
    """

    name: Literal[
        "Pyro Elemental Infusion",
        "Hydro Elemental Infusion",
        "Electro Elemental Infusion",
        "Cryo Elemental Infusion",
        "Anemo Elemental Infusion",
        "Geo Elemental Infusion",
        "Dendro Elemental Infusion",
    ]
    infused_elemental_type: DamageElementalType = DamageElementalType.PHYSICAL

    icon_type: Literal[
        IconType.ELEMENT_ENCHANT_ELEC,
        IconType.ELEMENT_ENCHANT_FIRE,
        IconType.ELEMENT_ENCHANT_WATER,
        IconType.ELEMENT_ENCHANT_ICE,
        IconType.ELEMENT_ENCHANT_WIND,
        IconType.ELEMENT_ENCHANT_ROCK,
        IconType.ELEMENT_ENCHANT_GRASS,
        IconType.ATK_UP,
    ] = IconType.ATK_UP

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.infused_elemental_type == DamageElementalType.PHYSICAL:
            # not set elemental type manually, get it from name
            element = self.name.split(" ")[0].upper()
            assert element in [
                "PYRO",
                "HYDRO",
                "ELECTRO",
                "CRYO",
                "ANEMO",
                "GEO",
                "DENDRO",
            ], "In ElementalInfusion: element not set right value"
            assert self.name.split(" ")[1:] == ["Elemental", "Infusion"]
            self.infused_elemental_type = DamageElementalType(element)
        if self.icon_type == IconType.ATK_UP:
            self.icon_type = ELEMENT_TO_ENCHANT_ICON[
                ElementType(self.infused_elemental_type.value)
            ]  # type: ignore

    def value_modifier_DAMAGE_ELEMENT_ENHANCE(
        self,
        value: DamageElementEnhanceValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> DamageElementEnhanceValue:
        """
        When self use skill, change physical to corresponding element.
        NOTE: this function will not change the usage.
        """
        assert mode == "REAL"
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, None
        ):
            # not corresponding character use skill, do nothing
            return value
        if self.usage <= 0:  # pragma: no cover
            # no usage, do nothing
            return value
        # change physical to other element
        if value.damage_elemental_type == DamageElementalType.PHYSICAL:
            value.damage_elemental_type = self.infused_elemental_type
        return value


class PrepareCharacterStatus(CharacterStatusBase):
    """
    Prepare status will trigger when self player action start, and try to
    use a specified skill, and skip next player action phase.
    if this character is stunned, it will do nothing.

    When this character is not active character, it will remove itself.
    """

    name: str
    version: str
    character_name: str
    skill_name: str

    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.SPECIAL] = IconType.SPECIAL

    done: bool = False

    def event_handler_SWITCH_CHARACTER(
        self, event: SwitchCharacterEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        If self switch character, after switch if this character is not active
        character, remove this status.
        """
        if event.action.player_idx != self.position.player_idx:
            # not self switch character
            return []
        return [
            RemoveObjectAction(
                object_position=self.position,
            )
        ]

    def event_handler_CHOOSE_CHARACTER(
        self, event: ChooseCharacterEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        As choose character only happens on character defeated, it is not
        possible.
        """
        if event.action.player_idx != self.position.player_idx:
            # not self choose character
            return []
        raise NotImplementedError("Should not be called")
        return [
            RemoveObjectAction(
                object_position=self.position,
            )
        ]

    def event_handler_PLAYER_ACTION_START(
        self, event: PlayerActionStartEventArguments, match: Any
    ) -> List[UseSkillAction]:
        """
        If self player action start, and this character is at front,
        use skill. it will also listen to Use Skill event to remove self, and skip
        player action.
        remove this status, and skip player action.
        """
        if event.player_idx != self.position.player_idx:
            # not self player action start, do nothing
            return []
        table = match.player_tables[self.position.player_idx]
        active_idx = table.active_character_idx
        if active_idx != self.position.character_idx:  # pragma: no cover
            # not active character, do nothing
            return []
        character = table.get_active_character()
        if character.is_stunned:
            # stunned, do nothing
            return []
        assert character.name == self.character_name
        for skill in character.skills:
            if skill.name == self.skill_name:
                self.done = True
                return [UseSkillAction(skill_position=skill.position)]
        else:
            raise AssertionError("Skill not found")

    def event_handler_USE_SKILL(
        self, event: UseSkillEventArguments, match: Any
    ) -> List[RemoveObjectAction | SkipPlayerActionAction]:
        """
        When use skill, and this status is marked as done, remove self and trigger
        skip player action.
        As it is a character status, it will always trigger after skill of character.
        """
        if self.done:
            return [
                # remove self
                RemoveObjectAction(object_position=self.position),
                # skip player action
                SkipPlayerActionAction(
                    action_label=PlayerActionLabels.SKILL,
                    do_combat_action=True,
                    position=event.action.skill_position,
                ),
            ]
        return []


class ReviveCharacterStatus(UsageCharacterStatus):
    name: str
    version: str
    heal: int

    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.REVIVE] = IconType.REVIVE

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[MakeDamageAction | RemoveObjectAction]:
        """
        when make damage, check whether this character's hp is 0. if so,
        heal it by self.heal hp.

        Using make damage action, so it will trigger immediately before
        receiving any other damage, and can be defeated by other damage.
        """
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        if character.hp > 0:
            # hp not 0, do nothing
            return []
        if self.usage <= 0:  # pragma: no cover
            # no usage, do nothing
            return []
        # heal this character by 1
        self.usage -= 1
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.HEAL,
                        target_position=character.position,
                        damage=-self.heal,
                        damage_elemental_type=DamageElementalType.HEAL,
                        cost=Cost(),
                    )
                ],
            )
        ] + self.check_should_remove()


class RoundEndAttackCharacterStatus(UsageCharacterStatus):
    """
    Make damage to this character at round end. It can also be heal, e.g.
    Pizza.
    """

    name: str
    version: str
    usage: int
    max_usage: int
    icon_type: Literal[IconType.DOT] = IconType.DOT
    damage: int
    damage_elemental_type: DamageElementalType

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        assert self.usage > 0
        self.usage -= 1
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        damage_type = DamageType.DAMAGE
        if self.damage < 0:
            damage_type = DamageType.HEAL
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=damage_type,
                        target_position=character.position,
                        damage=self.damage,
                        damage_elemental_type=self.damage_elemental_type,
                        cost=Cost(),
                    )
                ]
            )
        ]



from .agent_base import AgentBase
from ..server.match import Match
from ..server.interaction import (
    Responses,
    SwitchCardResponse,
    ChooseCharacterResponse,
    RerollDiceResponse,
    DeclareRoundEndResponse,
)


class NothingAgent(AgentBase):
    """
    Agent that do nothing, only response essential requests.
    """

    def generate_response(self, match: Match) -> Responses | None:
        for req in match.requests:
            if req.player_idx == self.player_idx:
                if req.name == "SwitchCardRequest":
                    return SwitchCardResponse(request=req, card_idxs=[])
                elif req.name == "ChooseCharacterRequest":
                    return ChooseCharacterResponse(
                        request=req, character_idx=req.available_character_idxs[0]
                    )
                elif req.name == "RerollDiceRequest":
                    return RerollDiceResponse(request=req, reroll_dice_idxs=[])
                elif req.name == "DeclareRoundEndRequest":
                    return DeclareRoundEndResponse(request=req)



from lpsim.agents.interaction_agent import InteractionAgent
from lpsim.server.match import Match, MatchState
from lpsim.server.deck import Deck
from tests.utils_for_test import (
    check_hp,
    get_random_state,
    get_test_id_from_command,
    make_respond,
    set_16_omni,
)


def test_collei():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "card 2 0 0 1 2 3",
            "sw_char 1 0",
            "sw_char 0 0",
            "skill 1 0 1 2",
            "sw_char 2 0",
            "TEST 2 p0 1 status",
            "skill 1 0 1 2",
            "sw_char 1 0",
            "end",
            "sw_char 0 0",
            "skill 0 0 1 2",
            "skill 1 0 1 2",
            "TEST 1 5 88 10 7 80 3",
            "TEST 3 p0 1 catalyzing field",
            "end",
            "skill 2 0 1 2",
            "skill 1 0 1 2",
            "sw_char 2 0",
            "skill 1 0 1 2",
            "TEST 1 3 88 10 7 66 3",
            "sw_char 0 0",
            "skill 1 0 1 2",
            "TEST 2 p0 0 status",
            "end",
            "skill 1 0 1 2",
            "end",
            "sw_char 1 0",
            "end",
        ],
        [
            "sw_card 0 1 2",
            "choose 0",
            "skill 1 0 1 2",
            "sw_char 2 0",
            "skill 1 0 1 2",
            "sw_char 0 0",
            "sw_char 2 0",
            "TEST 1 5 90 10 7 90 3",
            "TEST 3 p0 1 cayalyzing field",
            "sw_char 1 0",
            "end",
            "end",
            "end",
            "TEST 1 3 88 10 7 58 3",
            "end",
            "TEST 1 3 88 10 7 49 3",
            "sw_char 2 0",
            "sw_char 0 0",
            "card 2 0 0 1 2",
            "TEST 1 3 82 10 7 49 3",
            "skill 1 0 1 2",
            "TEST 2 p1 0 status",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Collei
        character:Nahida
        character:Fischl
        Floral Sidewinder*15
        """
    )
    # use old version cards
    old = {"name": "Floral Sidewinder", "version": "3.3"}
    deck_dict = deck.dict()
    deck_dict["cards"] += [old] * 15
    deck = Deck(**deck_dict)
    # change HP
    # for character in deck.characters:
    #     character.hp = 2
    #     character.max_hp = 2
    deck.characters[1].hp = 90
    deck.characters[1].max_hp = 90
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                cmd = cmd.split()
                pnum = int(cmd[2][1])
                snum = int(cmd[3])
                status = match.player_tables[pnum].team_status
                assert len(status) == snum
            elif test_id == 3:
                assert len(match.player_tables[0].team_status) == 1
                assert match.player_tables[0].team_status[0].usage == 1
                assert match.player_tables[0].team_status[0].name == "Catalyzing Field"
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == "__main__":
    test_collei()



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import AttackerSummonBase

from ...modifiable_values import DamageIncreaseValue

from ...action import Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    FactionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Summons


class TalismanSpirit_3_7(AttackerSummonBase):
    name: Literal["Talisman Spirit"] = "Talisman Spirit"
    version: Literal["3.7"] = "3.7"
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.CRYO
    damage: int = 1

    def value_modifier_DAMAGE_INCREASE(
        self,
        value: DamageIncreaseValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> DamageIncreaseValue:
        """
        When enemy receives physical or cryo damage, regardless of source,
        increase damage by 1
        """
        if value.damage_type != DamageType.DAMAGE:
            # not damage, do nothing
            return value
        if value.target_position.player_idx == self.position.player_idx:
            # attack self, no nothing
            return value
        if value.damage_elemental_type not in [
            DamageElementalType.CRYO,
            DamageElementalType.PHYSICAL,
        ]:
            # not cryo or physical, do nothing
            return value
        # increase damage
        assert mode == "REAL"
        value.damage += 1
        return value


# Skills


class SpringSpiritSummoning(ElementalSkillBase):
    name: Literal["Spring Spirit Summoning"] = "Spring Spirit Summoning"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(elemental_dice_color=DieColor.CRYO, elemental_dice_number=3)
    max_usage: int = 3

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        args = {
            "usage": self.max_usage,
            "max_usage": self.max_usage,
        }
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        if character.talent is not None:
            args.update(
                {
                    "talent_usage": 1,
                    "talent_max_usage": 1,
                }
            )
        return super().get_actions(match) + [
            self.create_team_status("Icy Quill", args),
        ]


class DivineMaidensDeliverance(ElementalBurstBase):
    name: Literal["Divine Maiden's Deliverance"] = "Divine Maiden's Deliverance"
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.CRYO, elemental_dice_number=3, charge=2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [self.create_summon("Talisman Spirit")]


# Talents


class MysticalAbandon_3_7(SkillTalent):
    name: Literal["Mystical Abandon"]
    version: Literal["3.7"] = "3.7"
    character_name: Literal["Shenhe"] = "Shenhe"
    cost: Cost = Cost(elemental_dice_color=DieColor.CRYO, elemental_dice_number=3)
    skill: Literal["Spring Spirit Summoning"] = "Spring Spirit Summoning"


# character base


class Shenhe_3_7(CharacterBase):
    name: Literal["Shenhe"]
    version: Literal["3.7"] = "3.7"
    element: ElementType = ElementType.CRYO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | SpringSpiritSummoning | DivineMaidensDeliverance
    ] = []
    faction: List[FactionType] = [FactionType.LIYUE]
    weapon_type: WeaponType = WeaponType.POLEARM

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Dawnstar Piercer",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            SpringSpiritSummoning(),
            DivineMaidensDeliverance(),
        ]


register_class(Shenhe_3_7 | MysticalAbandon_3_7 | TalismanSpirit_3_7)



from typing import Dict, List, Literal

from .....utils.class_registry import register_class

from ....modifiable_values import CostValue
from ....match import Match
from ....status.team_status.base import RoundTeamStatus
from ....consts import CostLabels, IconType, ObjectPositionType
from ....action import CreateObjectAction
from ....struct import Cost, ObjectPosition
from ....object_base import EventCardBase
from .....utils.desc_registry import DescDictType


class FallsAndFortuneStatus_4_3(RoundTeamStatus):
    name: Literal["Falls and Fortune"] = "Falls and Fortune"
    version: Literal["4.3"] = "4.3"
    usage: int = 1
    max_usage: int = 1
    icon_type: IconType = IconType.DEBUFF

    def value_modifier_COST(
        self,
        value: CostValue,
        match: Match,
        mode: Literal["TEST", "REAL"],
    ) -> CostValue:
        """
        if talent activated, increase switch cost
        """
        if (
            self.position.player_idx == value.position.player_idx
            and value.cost.label & CostLabels.SWITCH_CHARACTER.value != 0
        ):
            # increase switch cost regardless of player
            value.cost.any_dice_number += 1
        return value


class FallsAndFortune_4_3(EventCardBase):
    name: Literal["Falls and Fortune"]
    version: Literal["4.3"] = "4.3"
    cost: Cost = Cost(same_dice_number=1)

    def is_valid(self, match: Match) -> bool:
        return (
            len(match.player_tables[self.position.player_idx].dice.colors) >= 8
            and not match.player_tables[1 - self.position.player_idx].has_round_ended
        )

    def get_targets(self, match: Match) -> List[ObjectPosition]:
        # no target
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Match
    ) -> List[CreateObjectAction]:
        assert target is None
        return [
            CreateObjectAction(
                object_name=self.name,
                object_position=ObjectPosition(
                    player_idx=self.position.player_idx,
                    area=ObjectPositionType.TEAM_STATUS,
                    id=-1,
                ),
                object_arguments={},
            ),
            CreateObjectAction(
                object_name=self.name,
                object_position=ObjectPosition(
                    player_idx=1 - self.position.player_idx,
                    area=ObjectPositionType.TEAM_STATUS,
                    id=-1,
                ),
                object_arguments={},
            ),
        ]


desc: Dict[str, DescDictType] = {
    "TEAM_STATUS/Falls and Fortune": {
        "names": {"en-US": "Falls and Fortune", "zh-CN": "坍陷与契机"},
        "descs": {
            "4.3": {
                "en-US": "In this Round, both sides must spend 1 extra Elemental Die to switch characters.",  # noqa: E501
                "zh-CN": "本回合中，双方牌手进行「切换角色」行动时需要额外花费1个元素骰。",  # noqa: E501
            }
        },
    },
    "CARD/Falls and Fortune": {
        "names": {"en-US": "Falls and Fortune", "zh-CN": "坍陷与契机"},
        "descs": {
            "4.3": {
                "en-US": "Can only be played when you have at least 8 Elemental Dice remaining and your opponent has not ended their Round: In this Round, both sides must spend 1 extra Elemental Die to switch characters.",  # noqa: E501
                "zh-CN": "我方至少剩余8个元素骰，且对方未宣布结束时，才能打出：本回合中，双方牌手进行「切换角色」行动时需要额外花费1个元素骰。",  # noqa: E501
            }
        },
        "image_path": "cardface/Event_Event_Tanta.png",  # noqa: E501
        "id": 332026,
    },
}


register_class(FallsAndFortune_4_3 | FallsAndFortuneStatus_4_3, desc)



from typing import Any, List, Literal

from ....utils.class_registry import register_class


from ...summon.base import AttackerSummonBase

from ...action import ActionTypes, Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Summons


class Guoba_3_3(AttackerSummonBase):
    name: Literal["Guoba"] = "Guoba"
    version: Literal["3.3"] = "3.3"
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.PYRO
    damage: int = 2


# Skills


class DoughFu(PhysicalNormalAttackBase):
    name: Literal["Dough-Fu"] = "Dough-Fu"
    cost: Cost = PhysicalNormalAttackBase.get_cost(ElementType.PYRO)


class GuobaAttack(ElementalSkillBase):
    name: Literal["Guoba Attack"] = "Guoba Attack"
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(elemental_dice_color=DieColor.PYRO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        if self.is_talent_equipped(match):
            ret = super().get_actions(match)
            damage_action = ret[0]
            assert damage_action.type == ActionTypes.MAKE_DAMAGE
            ret.append(self.create_summon("Guoba"))
        else:
            ret: List[Actions] = [self.create_summon("Guoba"), self.charge_self(1)]
        return ret


class Pyronado(ElementalBurstBase):
    name: Literal["Pyronado"] = "Pyronado"
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.PYRO, elemental_dice_number=4, charge=2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [self.create_team_status("Pyronado")]


# Talents


class Crossfire_4_2(SkillTalent):
    version: Literal["4.2"] = "4.2"
    cost: Cost = Cost(elemental_dice_color=DieColor.PYRO, elemental_dice_number=3)
    name: Literal["Crossfire"]
    character_name: Literal["Xiangling"] = "Xiangling"
    skill: Literal["Guoba Attack"] = "Guoba Attack"


# character base


class Xiangling_3_8(CharacterBase):
    name: Literal["Xiangling"]
    version: Literal["3.8"] = "3.8"
    element: ElementType = ElementType.PYRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[DoughFu | GuobaAttack | Pyronado] = []
    faction: List[FactionType] = [FactionType.LIYUE]
    weapon_type: WeaponType = WeaponType.POLEARM

    def _init_skills(self) -> None:
        self.skills = [
            DoughFu(),
            GuobaAttack(),
            Pyronado(),
        ]


register_class(Xiangling_3_8 | Crossfire_4_2 | Guoba_3_3)



from typing import Any, Dict, List, Literal

from ....status.character_status.base import CharacterStatusBase
from ....action import (
    Actions,
    ChangeObjectUsageAction,
    CreateObjectAction,
    DrawCardAction,
    MakeDamageAction,
    RemoveObjectAction,
)
from .....utils.class_registry import register_class
from ....summon.base import AttackerSummonBase
from ....status.team_status.base import TeamStatusBase
from ....event import (
    GameStartEventArguments,
    ReceiveDamageEventArguments,
    RoundEndEventArguments,
    RoundPrepareEventArguments,
    SkillEndEventArguments,
)
from ....player_table import PlayerTable
from ....modifiable_values import DamageIncreaseValue, DamageValue
from ....match import Match
from ....struct import Cost, ObjectPosition
from ....character.character_base import (
    CharacterBase,
    ElementalBurstBase,
    ElementalNormalAttackBase,
    ElementalSkillBase,
    PassiveSkillBase,
    TalentBase,
)
from ....consts import (
    ELEMENT_TO_DAMAGE_TYPE,
    CostLabels,
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    FactionType,
    IconType,
    ObjectPositionType,
    ObjectType,
    WeaponType,
)
from .....utils.desc_registry import DescDictType


class LightningStrikeProbe_4_3(TeamStatusBase):
    name: Literal["Lightning Strike Probe"] = "Lightning Strike Probe"
    version: Literal["4.3"] = "4.3"
    usage: int = 1
    max_usage: int = 1
    icon_type: IconType = IconType.OTHERS

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Match
    ) -> List[Actions]:
        """
        When round prepare, reset usage
        """
        self.usage = self.max_usage
        return []

    def event_handler_SKILL_END(
        self, skill: SkillEndEventArguments, match: Match
    ) -> List[CreateObjectAction | RemoveObjectAction]:
        """
        When skill end, if skill user is on this side, attach lightning rod to
        active character.
        """
        if self.usage == 0:
            # no usage left, do nothing
            return []
        if skill.action.position.player_idx != self.position.player_idx:
            # not on this side, do nothing
            return []
        ret: List[CreateObjectAction | RemoveObjectAction] = []
        # check if any character has lightning rod attached
        characters = match.player_tables[self.position.player_idx].characters
        for character in characters:
            for status in character.status:
                if status.name == "Lightning Rod":
                    # lightning rod already attached, remove
                    ret.append(RemoveObjectAction(object_position=status.position))
        # attach lightning rod to the character
        self.usage -= 1
        ret.append(
            CreateObjectAction(
                object_name="Lightning Rod",
                object_position=skill.action.position.set_area(
                    ObjectPositionType.CHARACTER_STATUS
                ),
                object_arguments={"version": self.version},
            )
        )
        return ret


class LightningRod_4_3(CharacterStatusBase):
    name: Literal["Lightning Rod"] = "Lightning Rod"
    version: Literal["4.3"] = "4.3"
    usage: int = 1
    max_usage: int = 1
    icon_type: IconType = IconType.DEBUFF

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Match, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        """
        Opponent character with lightning rod attached take damage, increase
        damage by 1.
        """
        if self.usage == 0:
            # no usage left, do nothing
            return value
        if not value.is_corresponding_character_receive_damage(
            self.position, match, ignore_piercing=False
        ):
            # not corresponding character, do nothing
            return value
        if value.position.player_idx == self.position.player_idx:
            # damage from self, do nothing
            return value
        source = match.get_object(value.position)
        if source is not None and source.type == ObjectType.SKILL:
            # is skill, find corresponding character
            source = match.player_tables[value.position.player_idx].characters[
                value.position.character_idx
            ]
        if source is None or source.name not in [
            "Thunder Manifestation",
            "Thundering Shackles",
        ]:
            # not from thunder manifestation or its summon, do nothing
            return value
        # increase damage by 1
        value.damage += 1
        assert mode == "REAL"
        self.usage -= 1
        return value


def get_lightning_rod_target_idx(table: PlayerTable) -> int:
    """
    find character idx from table that has lightning rod attached. If no such
    character, return active character idx.
    """
    target: int | None = None
    for cid, character in enumerate(table.characters):
        for status in character.status:
            if status.name == "Lightning Rod":
                target = cid
                break
        if target is not None:
            break
    if target is None:
        target = table.active_character_idx
    return target


class ThunderingShacklesSummon_4_3(AttackerSummonBase):
    name: Literal["Thundering Shackles"] = "Thundering Shackles"
    version: Literal["4.3"] = "4.3"
    usage: int = 1
    max_usage: int = 1
    damage: int = 3
    damage_elemental_type: DamageElementalType = DamageElementalType.ELECTRO

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction | ChangeObjectUsageAction]:
        """
        When round end, make damage to right target, and decrease self usage.
        """
        assert self.usage > 0
        damage_type = DamageType.DAMAGE
        table = match.player_tables[1 - self.position.player_idx]
        target = get_lightning_rod_target_idx(table)
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=damage_type,
                        target_position=table.characters[target].position,
                        damage=self.damage,
                        damage_elemental_type=self.damage_elemental_type,
                        cost=Cost(),
                    )
                ],
            ),
            ChangeObjectUsageAction(object_position=self.position, change_usage=-1),
        ]


class StrifefulLightning(ElementalSkillBase):
    name: Literal["Strifeful Lightning"] = "Strifeful Lightning"
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ELECTRO,
        elemental_dice_number=3,
    )

    def get_actions(self, match: Match) -> List[Actions]:
        """
        Attack target with Lightning Rod attached.
        """
        table = match.player_tables[1 - self.position.player_idx]
        target = get_lightning_rod_target_idx(table)
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.DAMAGE,
                        target_position=table.characters[target].position,
                        damage=self.damage,
                        damage_elemental_type=self.damage_type,
                        cost=self.cost.copy(),
                    )
                ],
            ),
            self.charge_self(1),
        ]


class ThunderingShackles(ElementalBurstBase):
    name: Literal["Thundering Shackles"] = "Thundering Shackles"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ELECTRO,
        elemental_dice_number=3,
        charge=2,
    )

    def get_actions(self, match: Match) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_summon("Thundering Shackles"),
        ]


class LightningProbe(PassiveSkillBase):
    name: Literal["Lightning Probe"] = "Lightning Probe"
    version: Literal["4.3"] = "4.3"

    def event_handler_GAME_START(
        self, event: GameStartEventArguments, match: Match
    ) -> List[CreateObjectAction]:
        """
        When game start, create Lightning Strike Probe on opponent's side.
        """
        return [
            CreateObjectAction(
                object_name="Lightning Strike Probe",
                object_position=ObjectPosition(
                    player_idx=1 - self.position.player_idx,
                    area=ObjectPositionType.TEAM_STATUS,
                    id=-1,
                ),
                object_arguments={"version": self.version},
            )
        ]


class GrievingEcho_4_3(TalentBase):
    name: Literal["Grieving Echo"]
    version: Literal["4.3"] = "4.3"
    character_name: Literal["Thunder Manifestation"] = "Thunder Manifestation"
    cost_label: int = (
        CostLabels.CARD.value | CostLabels.TALENT.value | CostLabels.EQUIPMENT.value
    )
    cost: Cost = Cost()
    usage: int = 1

    def is_valid(self, match: Any) -> bool:
        if self.position.area != ObjectPositionType.HAND:
            # not in hand, cannot equip
            raise AssertionError("Talent is not in hand")
        return len(self.get_targets(match)) > 0

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """
        All ayaka are targets
        """
        assert self.position.area != ObjectPositionType.CHARACTER
        ret: List[ObjectPosition] = []
        for c in match.player_tables[self.position.player_idx].characters:
            if c.name == self.character_name and c.is_alive:
                ret.append(c.position)
        return ret

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Match
    ) -> List[Actions]:
        """
        When round prepare, reset usage
        """
        self.usage = 1
        return []

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Match
    ) -> List[DrawCardAction]:
        """
        Opponent character with lightning rod attached take damage, draw card.
        """
        target_position = event.final_damage.target_position
        if self.position.area != ObjectPositionType.CHARACTER:
            # not equipped, do nothing
            return []
        if self.usage <= 0:
            # no usage left, do nothing
            return []
        if target_position.player_idx == self.position.player_idx:
            # damage to self, do nothing
            return []
        target_character: CharacterBase = match.get_object(target_position)  # type: ignore
        for status in target_character.status:
            if status.name == "Lightning Rod":
                # opposite character has lightning rod attached, draw card
                self.usage = 0
                return [
                    DrawCardAction(
                        player_idx=self.position.player_idx,
                        number=1,
                        draw_if_filtered_not_enough=True,
                    )
                ]
        return []


class ThunderManifestation_4_3(CharacterBase):
    name: Literal["Thunder Manifestation"]
    version: Literal["4.3"] = "4.3"
    element: ElementType = ElementType.ELECTRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        ElementalNormalAttackBase
        | StrifefulLightning
        | ThunderingShackles
        | LightningProbe
    ] = []
    faction: List[FactionType] = [FactionType.MONSTER]
    weapon_type: WeaponType = WeaponType.OTHER

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name="Thunderous Wingslash",
                damage_type=ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost=ElementalNormalAttackBase.get_cost(self.element),
            ),
            StrifefulLightning(),
            ThunderingShackles(),
            LightningProbe(),
        ]


desc: Dict[str, DescDictType] = {
    "CHARACTER/Thunder Manifestation": {
        "names": {"en-US": "Thunder Manifestation", "zh-CN": "雷音权现"},
        "image_path": "cardface/Char_Monster_Raijin.png",  # noqa: E501
        "id": 2402,
        "descs": {"4.3": {"en-US": "", "zh-CN": ""}},
    },
    "SKILL_Thunder Manifestation_NORMAL_ATTACK/Thunderous Wingslash": {
        "names": {"en-US": "Thunderous Wingslash", "zh-CN": "轰霆翼斩"},
        "descs": {
            "4.3": {"en-US": "Deals 1 Electro DMG.", "zh-CN": "造成1点雷元素伤害。"}
        },
    },
    "SKILL_Thunder Manifestation_ELEMENTAL_SKILL/Strifeful Lightning": {
        "names": {"en-US": "Strifeful Lightning", "zh-CN": "雷墙倾轧"},
        "descs": {
            "4.3": {
                "en-US": "Deals 3 Electro DMG to opposing characters affected by Lightning Rod. (If there are no eligible opposing characters, deals DMG to the active character instead)",  # noqa: E501
                "zh-CN": "对附属有雷鸣探知的敌方角色造成3点雷元素伤害。（如果敌方不存在符合条件角色，则改为对出战角色造成伤害）",  # noqa: E501
            }
        },
    },
    "CHARACTER_STATUS/Lightning Rod": {
        "names": {"en-US": "Lightning Rod", "zh-CN": "雷鸣探知"},
        "descs": {
            "4.3": {
                "en-US": "While this status is active, can be triggered once: DMG received by the attached character from Thunder Manifestation or its summons is increased by 1.\n(Only one of this status can exist on the field at once. Some of Thunder Manifestation's skills will target the character to which this is attached.)",  # noqa: E501
                "zh-CN": "此状态存在期间，可以触发1次：所附属角色受到雷音权现及其召唤物造成的伤害+1。\n（同一方场上最多存在一个此状态。雷音权现的部分技能，会以所附属角色为目标。）",  # noqa: E501
            }
        },
    },
    "SKILL_Thunder Manifestation_ELEMENTAL_BURST/Thundering Shackles": {
        "names": {"en-US": "Thundering Shackles", "zh-CN": "轰雷禁锢"},
        "descs": {
            "4.3": {
                "en-US": "Deals 2 Electro DMG, summons 1 Thundering Shackles.",
                "zh-CN": "造成2点雷元素伤害，召唤轰雷禁锢。",
            }
        },
    },
    "SUMMON/Thundering Shackles": {
        "names": {"en-US": "Thundering Shackles", "zh-CN": "轰雷禁锢"},
        "descs": {
            "4.3": {
                "en-US": "End Phase: Deals 3 Electro DMG to opposing characters affected by Lightning Rod. (If there are no eligible opposing characters, deals DMG to the active character instead)\nUsage(s): 1",  # noqa: E501
                "zh-CN": "结束阶段：对附属有雷鸣探知的敌方角色造成造成3点雷元素伤害。（如果敌方不存在符合条件角色，则改为对出战角色造成伤害）\n可用次数：1",  # noqa: E501
            }
        },
        "image_path": "cardface/Summon_Raijin.png",  # noqa: E501
    },
    "SKILL_Thunder Manifestation_PASSIVE/Lightning Probe": {
        "names": {"en-US": "Lightning Probe", "zh-CN": "雷霆探知"},
        "descs": {
            "4.3": {
                "en-US": "(Passive) When battle begins, create a Lightning Strike Probe on the opponent's side of the field.",  # noqa: E501
                "zh-CN": "【被动】战斗开始时，在敌方场上生成雷霆探针。",
            }
        },
    },
    "TEAM_STATUS/Lightning Strike Probe": {
        "names": {"en-US": "Lightning Strike Probe", "zh-CN": "雷霆探针"},
        "descs": {
            "4.3": {
                "en-US": "After a character on whose side of the field this card is on uses a Skill: Attach Lightning Rod to the active character on that side. (Once per Round)",  # noqa: E501
                "zh-CN": "所在阵营角色使用技能后：对所在阵营出战角色附属雷鸣探知。（每回合1次）",
            }
        },
        "image_path": "status/Debuff_Raijin_S.png",
    },
    "TALENT_Thunder Manifestation/Grieving Echo": {
        "names": {"en-US": "Grieving Echo", "zh-CN": "悲号回唱"},
        "descs": {
            "4.3": {
                "en-US": "When Thunder Manifestation, with this card equipped, is in play, when opponents with Lightning Rod attached take DMG: You draw 1 card (Once per Round)\n(Your deck must contain Thunder Manifestation to add this card to your deck)",  # noqa: E501
                "zh-CN": "装备有此牌的雷音权现在场，附属有雷鸣探知的敌方角色受到伤害时：我方抓1张牌。（每回合1次）\n（牌组中包含雷音权现，才能加入牌组）",  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Talent_Raijin.png",  # noqa: E501
        "id": 224021,
    },
}


register_class(
    LightningRod_4_3
    | LightningStrikeProbe_4_3
    | ThunderingShacklesSummon_4_3
    | GrievingEcho_4_3
    | ThunderManifestation_4_3,
    desc,
)



from typing import Any, List, Literal

from ....utils.class_registry import register_base_class, register_class

from ...modifiable_values import CombatActionValue, CostValue


from ...action import Actions, CreateDiceAction, DrawCardAction, RemoveObjectAction

from ...event import (
    PlayerActionStartEventArguments,
    ReceiveDamageEventArguments,
    RoundPrepareEventArguments,
    SkillEndEventArguments,
    SwitchCharacterEventArguments,
    UseCardEventArguments,
)

from ...struct import Cost
from ...consts import (
    CostLabels,
    DamageElementalType,
    DamageType,
    IconType,
    ObjectPositionType,
    PlayerActionLabels,
)
from .base import RoundEffectSupportBase, SupportBase


class ItemBase(SupportBase):
    cost_label: int = CostLabels.CARD.value | CostLabels.ITEM.value


register_base_class(ItemBase)


class RoundEffectItemBase(RoundEffectSupportBase, ItemBase):
    cost_label: int = CostLabels.CARD.value | CostLabels.ITEM.value


class ParametricTransformer_3_3(ItemBase):
    name: Literal["Parametric Transformer"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(any_dice_number=2)
    usage: int = 0
    max_usage: int = 3
    create_dice_number: int = 3
    progress_got: bool = False
    icon_type: Literal[IconType.COUNTER] = IconType.COUNTER

    def event_handler_PLAYER_ACTION_START(
        self, event: PlayerActionStartEventArguments, match: Any
    ) -> List[Actions]:
        """
        When anyone start action, reset progress_got.
        """
        self.progress_got = False
        return []

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Any
    ) -> List[Actions]:
        """
        If elemental damage, gain progress
        """
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return []
        if event.final_damage.damage_type == DamageType.HEAL:
            # heal, do nothing
            return []
        if event.final_damage.damage_elemental_type in [
            DamageElementalType.PHYSICAL,
            DamageElementalType.PIERCING,
            DamageElementalType.HEAL,
        ]:
            # physical or piercing or heal, do nothing
            return []
        # get progress
        self.progress_got = True
        return []

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[CreateDiceAction | RemoveObjectAction]:
        """
        If progress got, increase usage. If usage receives max_usage,
        remove self and create different dice.
        """
        if self.progress_got:
            self.usage += 1
        self.progress_got = False
        if self.usage == self.max_usage:
            # remove self and create different dice
            return [
                RemoveObjectAction(
                    object_position=self.position,
                ),
                CreateDiceAction(
                    player_idx=self.position.player_idx,
                    number=self.create_dice_number,
                    different=True,
                ),
            ]
        return []


class NRE_4_1(RoundEffectItemBase):
    name: Literal["NRE"]
    version: Literal["4.1"] = "4.1"
    cost: Cost = Cost(same_dice_number=1)
    max_usage_per_round: int = 1

    def draw_food_card(self) -> DrawCardAction:
        return DrawCardAction(
            player_idx=self.position.player_idx,
            number=1,
            draw_if_filtered_not_enough=False,
            whitelist_cost_labels=CostLabels.FOOD.value,
        )

    def play(self, match: Any) -> List[Actions]:
        return [self.draw_food_card()]

    def event_handler_USE_CARD(
        self, event: UseCardEventArguments, match: Any
    ) -> List[Actions]:
        if self.position.area == ObjectPositionType.SUPPORT:
            # on support area, do effect
            if self.usage <= 0:
                # usage is 0, do nothing
                return []
            if event.action.card_position.player_idx != self.position.player_idx:
                # not our character use card, do nothing
                return []
            if event.card_cost.label & CostLabels.FOOD.value == 0:
                # not food card, do nothing
                return []
            # our use food card, draw new food card
            self.usage -= 1
            return [self.draw_food_card()]
        # otherwise, do normal response
        return super().event_handler_USE_CARD(event, match)


class NRE_3_3(NRE_4_1):
    version: Literal["3.3"]
    cost: Cost = Cost(any_dice_number=2)


class RedFeatherFan_3_7(RoundEffectItemBase):
    name: Literal["Red Feather Fan"]
    version: Literal["3.7"] = "3.7"
    cost: Cost = Cost(same_dice_number=2)
    max_usage_per_round: int = 1
    switch_count: int = 0  # activate after first switch

    # when decrease cost, should consume usage regardless of whether fast
    # action success
    is_cost_decreased: bool = False

    def play(self, match: Any) -> List[Actions]:
        self.switch_count = 0
        self.is_cost_decreased = False
        return super().play(match)

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        self.switch_count = 0
        self.is_cost_decreased = False
        return super().event_handler_ROUND_PREPARE(event, match)

    def event_handler_SWITCH_CHARACTER(
        self, event: SwitchCharacterEventArguments, match: Any
    ) -> List[Actions]:
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return []
        if self.position.player_idx != event.action.player_idx:
            # not our character switch, do nothing
            return []
        # activate
        self.switch_count += 1
        return []

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> CostValue:
        """
        if activated, and our switch character, try to decrease cost.
        if success, mark is_cost_decreased.
        """
        if (
            self.switch_count < 1  # for cost, ignore first switch
            or self.position.area != ObjectPositionType.SUPPORT
            or self.position.player_idx != value.position.player_idx
            or self.usage <= 0
            or value.cost.label & CostLabels.SWITCH_CHARACTER.value == 0
        ):
            # not activated, or not equipped, or not our character,
            # or no usage, or not switch character, do nothing
            return value
        # decrease cost
        if value.cost.decrease_cost(None):
            if mode == "REAL":
                self.is_cost_decreased = True
        return value

    def value_modifier_COMBAT_ACTION(
        self, value: CombatActionValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> CombatActionValue:
        """
        If activated, and our switch character, try to act as fast action.
        Then if change to fast or decrease cost, consume usage.
        """
        if (
            self.switch_count < 2  # for combat action, ignore first two switch
            or self.position.area != ObjectPositionType.SUPPORT
            or self.position.player_idx != value.position.player_idx
            or self.usage <= 0
            or value.action_label & PlayerActionLabels.SWITCH.value == 0
        ):
            # not activated, or not equipped, or not our character,
            # or no usage, or not switch character, do nothing
            return value
        # change to fast action
        changed = False
        if value.do_combat_action:
            value.do_combat_action = False
            changed = True
        assert mode == "REAL"
        # decrease usage
        if changed or self.is_cost_decreased:
            self.usage -= 1
        return value


class TreasureSeekingSeelie_3_7(ItemBase):
    name: Literal["Treasure-Seeking Seelie"]
    version: Literal["3.7"] = "3.7"
    cost: Cost = Cost(same_dice_number=1)
    usage: int = 0
    max_usage: int = 3
    icon_type: Literal[IconType.COUNTER] = IconType.COUNTER

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[RemoveObjectAction | DrawCardAction]:
        """
        If our character use any skill, increase usage, and if usage
        reaches max_usage, draw 3 cards and remove self.
        """
        if not self.position.check_position_valid(
            event.action.position,
            match,
            player_idx_same=True,
            source_area=ObjectPositionType.SUPPORT,
            target_area=ObjectPositionType.SKILL,
        ):
            # not on support area, or not our character use skill
            return []
        # add usage
        self.usage += 1
        if self.usage != self.max_usage:
            # not reach max usage, do nothing
            return []
        # reach max usage, draw 3 cards and remove self
        return [
            DrawCardAction(
                player_idx=self.position.player_idx,
                number=3,
                draw_if_filtered_not_enough=True,
            ),
            RemoveObjectAction(object_position=self.position),
        ]


register_class(
    ParametricTransformer_3_3
    | NRE_4_1
    | NRE_3_3
    | RedFeatherFan_3_7
    | TreasureSeekingSeelie_3_7
)



from typing import Any, Dict, List, Literal, Set, Type

from pydantic import PrivateAttr

from ....action import ActionTypes, Actions, CreateRandomObjectAction
from .....utils.desc_registry import DescDictType
from .....utils.class_registry import (
    get_class_list_by_base_class,
    get_instance,
    register_class,
)
from ....card.support.locations import LocationBase
from ....match import Match
from ....consts import IconType, ObjectPositionType
from ....event import GameStartEventArguments, UseCardEventArguments
from ....card.support.items import ItemBase
from ....card.event.foods import FoodCardBase
from ....struct import Cost, ObjectPosition
from ....card.support.base import UsageWithRoundRestrictionSupportBase
from ....card.support.companions import CompanionBase


class Mamere_4_3(CompanionBase, UsageWithRoundRestrictionSupportBase):
    name: Literal["Mamere"]
    version: Literal["4.3"] = "4.3"
    cost: Cost = Cost()
    max_usage_one_round: int = 1
    usage: int = 3

    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    _accept_card_types: Type = PrivateAttr(
        LocationBase | FoodCardBase | ItemBase | CompanionBase
    )
    _accept_class_list: List[str] | None = PrivateAttr(None)
    _accept_class_name_set: Set[str] = PrivateAttr(set())
    _accept_version: str | None = PrivateAttr(None)

    available_handler_in_deck: List[ActionTypes] = [ActionTypes.GAME_START]

    def _get_candidate_list(self, match: Match) -> List[str]:
        if self._accept_class_list is not None:
            return self._accept_class_list
        kwargs: Dict[str, Any] = {"exclude": set([self.name])}
        default_version = match.player_tables[
            self.position.player_idx
        ].player_deck_information.default_version
        if default_version is not None:
            kwargs["version"] = default_version
        self._accept_version = default_version
        candidate_list = get_class_list_by_base_class(self._accept_card_types, **kwargs)
        self._accept_class_list = candidate_list
        self._accept_class_name_set = set(candidate_list)
        return candidate_list

    def event_handler_GAME_START(
        self, event: GameStartEventArguments, match: Match
    ) -> List[Actions]:
        """
        check all cards that able to generate. If they contains GAME_START
        event and available in deck, call them.
        """
        candidate_list = self._get_candidate_list(match)
        args = {}
        if self._accept_version is not None:
            args["version"] = self._accept_version
        candidate_instance = [
            get_instance(self._accept_card_types, {"name": name, **args})
            for name in candidate_list
        ]
        res: List[Actions] = []
        for i in candidate_instance:
            if ActionTypes.GAME_START in i.available_handler_in_deck:
                res += i.event_handler_GAME_START(event, match)
        return res

    def event_handler_USE_CARD(
        self, event: UseCardEventArguments, match: Match
    ) -> List[Actions]:
        if self.position.area != ObjectPositionType.SUPPORT:
            # not placed to support, do original action
            return super().event_handler_USE_CARD(event, match)
        if not self.has_usage():
            # no usage, return
            return []
        if event.action.card_position.player_idx != self.position.player_idx:
            # not self use card, return
            return []
        card_name = event.card_name
        if card_name == "Mamere":
            # using Mamere, return
            return []
        # recall _get_candidate_list to create data
        candidate_list = self._get_candidate_list(match)
        if card_name in self._accept_class_name_set:
            # use, and generate new card
            self.use()
            res: List[Actions] = []
            position = ObjectPosition(
                player_idx=self.position.player_idx, area=ObjectPositionType.HAND, id=-1
            )
            args = {}
            if self._accept_version is not None:
                args["version"] = self._accept_version
            res.append(
                CreateRandomObjectAction(
                    object_names=candidate_list,
                    object_position=position,
                    object_arguments=args,
                    number=1,
                )
            )
            res += self.check_should_remove()
            return res
        return []


desc: Dict[str, DescDictType] = {
    "SUPPORT/Mamere": {
        "names": {"en-US": "Mamere", "zh-CN": "玛梅赫"},
        "descs": {
            "4.3": {
                "en-US": "After you play a Food/Location/Companion/Item Action Card other than Mamere: Randomly create 1 Food/Location/Companion/Item Action Card other than Mamere, and add it to your hand. (Once per Round).\nUsage(s): 3",  # noqa: E501
                "zh-CN": "我方打出「玛梅赫」以外的「料理」/「场地」/「伙伴」/「道具」行动牌后：随机生成1张「玛梅赫」以外的「料理」/「场地」/「伙伴」/「道具」行动牌，将其加入手牌。（每回合1次）\n可用次数：3",  # noqa: E501
            }
        },
        "image_path": "cardface/Assist_NPC_Mamere.png",  # noqa: E501
        "id": 322021,
    },
}


register_class(Mamere_4_3, desc)



from typing import Any, Dict, List, Literal
from lpsim.server.action import DrawCardAction
from lpsim.server.object_base import EventCardBase
from lpsim.server.struct import Cost, ObjectPosition
from lpsim.utils.class_registry import register_class
from lpsim.utils.desc_registry import DescDictType


class BigStrategize_3_3(EventCardBase):
    """
    Big Strategize
    """
    name: Literal['Big Strategize'] = 'Big Strategize'
    version: Literal['3.3'] = '3.3'
    cost = Cost(
        any_dice_number = 2
    )

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[DrawCardAction]:
        """
        Act the card. Draw three cards.
        """
        assert target is None  # no targets
        return [DrawCardAction(
            player_idx = self.position.player_idx, 
            number = 3,
            draw_if_filtered_not_enough = True
        )]


card_descs: Dict[str, DescDictType] = {
    'CARD/Big Strategize': {
        'names': {
            'zh-CN': '大心海',
            'en-US': 'Big Strategize',
        },
        "descs": {
            "3.3": {
                "zh-CN": "抓3张牌。",
                "en-US": "Draw 3 cards."
            }
        },
        "id": 332099,
        "image_path": "cardface/Event_Event_Yunchouweiwo.png"
    }
}


register_class(BigStrategize_3_3, card_descs)



"""
Tools that based on other modules.
"""
import json
from typing import List, Tuple

from .server.event_handler import OmnipotentGuideEventHandler_3_3

from .server.deck import Deck
from .server.match import Match, MatchConfig, MatchState
from .agents.interaction_agent import InteractionAgent


def read_log(
    log_str: str, use_16_omni: bool | None = None
) -> Tuple[List[InteractionAgent], Match]:
    """
    Read log and return agents and a match.

    Args:
        log_str (str): log string with JSON format. It must contain
            `start_deck`, `match_random_state` and `command_history`, and
            optionally `match_config`.
        use_16_omni (bool | None, optional): If True, use OmnipotentGuide.
            If False, not use it. If None, determined by dice number from
            match_config, if config exists and is 16 then True, otherwise
            False.
    """
    data = json.loads(log_str)
    if "version" not in data:
        data["version"] = "1.0"
    if data["version"] == "1.0":
        decks = []
        for d in data["start_deck"]:
            if isinstance(d, str):
                decks.append(Deck.from_str(d))
            else:
                decks.append(Deck(**d))
        match = Match(random_state=data["match_random_state"])
        if "match_config" in data:
            match.config = MatchConfig(**data["match_config"])
        if use_16_omni is None:
            if (
                "match_config" in data
                and data["match_config"]["initial_dice_number"] == 16
            ):
                use_16_omni = True
            else:
                use_16_omni = False
        if use_16_omni:
            match.event_handlers.append(OmnipotentGuideEventHandler_3_3())
        match.set_deck(decks)
        commands = data["command_history"]
        agents: List[InteractionAgent] = []
        if len(commands[0]) > 0:  # pragma: no cover
            if isinstance(commands[0][0], (list, tuple)):
                # new version, extract only string
                commands = [[y[1] for y in x] for x in commands]
        commands = [[y for y in x if not y.startswith("TEST")] for x in commands]
        for cnum, command in enumerate(commands):
            agent = InteractionAgent(
                player_idx=cnum, commands=command, only_use_command=True
            )
            agents.append(agent)
        return agents, match
    else:
        raise ValueError(f'Unknown log version: {data["version"]}')


def interact_match_with_agents(
    agent_0: InteractionAgent, agent_1: InteractionAgent, match: Match
) -> None:
    """
    Interact match with agents.
    """
    if match.state == MatchState.WAITING:
        assert match.start()[0]
    match.step()
    while len(agent_1.commands) > 0 or len(agent_0.commands) > 0:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # respond
        resp = agent.generate_response(match)
        assert resp is not None
        match.respond(resp)
        match.step()



import asyncio
import logging
import json
import uuid
import os
import time
import datetime
from typing import Literal, List
from fastapi import FastAPI, HTTPException, Request
from fastapi.responses import JSONResponse
from fastapi.middleware.cors import CORSMiddleware
from starlette.middleware.gzip import GZipMiddleware
from pydantic import BaseModel
import uvicorn

from ..utils.desc_registry import get_desc_patch

from ..server.match import Match, MatchConfig
from ..server.deck import Deck
from ..agents import InteractionAgent
from .utils import get_new_match
from ..utils.deck_code import deck_code_data
from .__version__ import __version_tuple__, __version__, __frontend_version__


class EndpointFilter(logging.Filter):
    """Filter class to exclude specific endpoints from log entries."""

    def __init__(self, excluded_endpoints: list[str]) -> None:
        """
        Initialize the EndpointFilter class.

        Args:
            excluded_endpoints (list[str]): A list of endpoints to be excluded from log
                entries.
        """
        self.excluded_endpoints = excluded_endpoints

    def filter(self, record: logging.LogRecord) -> bool:
        """
        Filter out log entries for excluded endpoints.

        Args:
            record (logging.LogRecord): The log record to be filtered.

        Returns:
            bool: True if the log entry should be included, False otherwise.
        """
        if record.args and len(record.args) >= 3:
            endpoint: str = record.args[2]  # type: ignore
            for excluded_endpoint in self.excluded_endpoints:
                if endpoint.startswith(excluded_endpoint):
                    return False
            return True
        return False


class ResetData(BaseModel):
    fixed_random_seed: bool = False
    offset: int = 0
    match_config: MatchConfig | None = None
    rich_mode: bool = False
    match_state: Match | None = None
    match_state_idx: int | None = None


class DeckData(BaseModel):
    deck_str: str
    player_idx: int


class RespondData(BaseModel):
    player_idx: int
    command: str
    uuid: str
    frame_number: int = -1


class HTTPServer:
    """
    Simple HTTP server based on HTTP API. No safety protocol and data
    visibility support now, i.e. everyone can get full information of opponent
    player and system states.

    To check APIs, use /docs or /redoc after run the server.
    """

    def __init__(
        self,
        allow_origins: List[str] = ["*"],
        decks: List[Deck | str] = [],
        match_config: MatchConfig | None = None,
        reset_log_save_path: str | None = None,
        room_name: str = "",
        excluded_log_endpoints: List[str] = ["/request", "/state"],
        get_state_timeout: float = 10,
        get_state_sleeptime: float = 0.5,
    ):
        """
        Init the HTTPServer instance.

        Args:
            allow_origins (List[str]): list of allowed origins, default is ['*']
            decks (List[Deck | str]): list of decks of players, each element can be a
                string or a Deck object. If a string, it will be parsed to a Deck
                object.
            match_config (MatchConfig): match config of the match, default is None,
                which means use default config.
            reset_log_save_path (str): if not None, save the log when reset match to the
                path. The log is encoded in json, and log name is the timestamp.
            room_name (str): name of the room. If not empty, only request with
                room=room_name will be accepted.
            excluded_log_endpoints (List[str]): endpoints that will not be logged.
                Default is ['/request', '/state'].
        """
        self.app = FastAPI()
        self.decks = [
            Deck.from_str(deck) if isinstance(deck, str) else deck for deck in decks
        ]
        self.reset_log_save_path = reset_log_save_path
        self.room_name = room_name
        self.excluded_log_endpoints = excluded_log_endpoints
        # timeout and sleeptime for long polling get state. If do not want to
        # perform long polling, set timeout to less than 0.
        self.get_state_timeout = get_state_timeout
        self.get_state_sleeptime = get_state_sleeptime
        logging.getLogger("uvicorn.access").addFilter(
            EndpointFilter(self.excluded_log_endpoints)
        )
        if match_config is not None and match_config.history_level < 10:
            raise ValueError("history_level must be at least 10 for HTTPServer")
        match, random_state = get_new_match(self.decks, match_config=match_config)
        self.uuid = str(uuid.uuid4())
        self.match = match
        self.match_random_state = random_state
        self.agent_0 = InteractionAgent(player_idx=0, only_use_command=True)
        self.agent_1 = InteractionAgent(player_idx=1, only_use_command=True)
        # used to record all uploaded deck codes to the server
        self.uploaded_deck_codes = []

        if len(self.decks) == 0:
            # no deck input at init, create two empty decks.
            self.decks = [Deck.from_str("""""") for _ in range(2)]

        # attrs to save history. command is saved separately for two players,
        # and each line contains [idx, command], idx is idx of last history.
        self.command_history = [[], []]
        self.start_deck = self.decks[:]

        app = self.app
        # Add the CORSMiddleware to the app
        app.add_middleware(
            CORSMiddleware,
            allow_origins=allow_origins,
            # allow_origins=['*'],
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )
        # Add the GZipMiddleware to the app
        app.add_middleware(GZipMiddleware)

        @app.middleware("http")
        async def check_room_name(request: Request, call_next):
            query_params = request.query_params
            if self.room_name != "" and (
                "room" not in query_params or query_params["room"] != self.room_name
            ):
                return JSONResponse(status_code=404, content="Room name wrong")
            response = await call_next(request)
            return response

        """API handlers for app"""

        @app.get("/version")
        async def get_version():
            """
            Return the version of lpsim.

            Return:
                version: version string
                version_tuple: version tuple
                support_version: support version of frontend
                info: information of the server
            """
            return {
                "version": __version__,
                "version_tuple": __version_tuple__,
                "support_version": __frontend_version__,
                "info": {
                    "class": "HTTPServer",
                },
            }

        @app.get("/patch")
        async def get_patch():
            """
            Return the desc patch.

            Return:
                version: version string
                patch: patch dict, which contains all descriptions with type
                    `Dict[str, DescDictType]`.
            """
            patch = get_desc_patch()
            return {"version": "1.0", "patch": patch}

        @app.post("/reset")
        async def post_reset(data: ResetData):
            """
            Reset the match.
            If match_state_idx is not None, the match will be reset to the
            state of match_state_idx in history. This will keep histories
            before match_state_idx.

            Otherwise, if match_state is not None, the match will be reset to
            match_state. This will clear all histories and start with zero.
            (Currently not implemented).

            Otherwise, a new match will be created, with
            fixed_random_seed, offset and match_config and rich_mode. Note when
            match_config is None, current match_config in self.match will be
            used.

            The return has same structure as state API.

            If it is a full match, four
            fields will be returned: uuid, idx, match and type. Here match is the
            full match state, which is parsed from `Match` directly.

            If it is a diff match, four fields will be returned: uuid, idx, match_diff
            and type. Here match_diff is the diff of the match state, which is
            generated by dictdiffer, to describe difference between two match
            states. Note that frontend show frames totally based on matches, so
            need to generate multiple diffs to show multiple changes (e.g. using skill
            may generate UseSkill, MakeDamage, DrawCard, etc. multiple frames).

            Return:
                uuid: uuid of the match
                idx: index of the match state
                match: full match state
                type: "FULL"
            """
            self.save_log(self.reset_log_save_file)
            match = self.match
            fixed_random_seed = data.fixed_random_seed
            if fixed_random_seed:
                raise NotImplementedError("fixed_random_seed not supported")
            match_config = data.match_config
            if match_config is None:
                match_config = match.config
            rich = data.rich_mode
            match_state = data.match_state
            match_state_idx = data.match_state_idx
            if match_state_idx is not None:
                try:
                    match = self.match.new_match_from_history(match_state_idx)
                except AssertionError as e:
                    raise HTTPException(status_code=404, detail=str(e))
                self.match = match
                # remove command history that after match_state_idx. No need to
                # change match random state and start deck.
                current_cmd_hist = self.command_history[:]
                self.command_history = [[], []]
                for source, target in zip(current_cmd_hist, self.command_history):
                    for idx, cmd, order in source:
                        if idx >= match_state_idx:
                            break
                        target.append([idx, cmd, order])
            elif match_state is not None:
                match = match_state
                match._save_history()
                match_state_idx = len(match._history_diff) - 1
                # when reset by match_state, cannot record history with only
                # match random state, set to None to notify that cannot save.
                self.command_history = [[], []]
                self.match_random_state = None
                self.start_deck = [
                    x.player_deck_information for x in match.player_tables
                ]
            else:
                self.match, self.match_random_state = get_new_match(
                    decks=self.decks,
                    rich_mode=rich,
                    match_config=match_config,
                )
                match_state_idx = 0
                match = self.match
                self.command_history = [[], []]
                self.start_deck = [
                    x.player_deck_information for x in match.player_tables
                ]
            self.uuid = str(uuid.uuid4())
            return {
                "uuid": self.uuid,
                "idx": match_state_idx,
                "match": match.dict(),
                "type": "FULL",
            }

        @app.get("/deck/{player_idx}")
        async def get_deck(player_idx: int):
            """
            get deck of a player.

            Args:
                player_idx (int): index of the player, 0 or 1.

            Return:
                deck: deck of the player, which directly parsed from `Deck` object.
            """
            if player_idx < 0 or player_idx > 1:
                raise HTTPException(status_code=404, detail="Player not found")
            return self.decks[player_idx].dict()

        @app.get("/decks")
        async def get_decks():
            """
            get decks of both players.

            Return:
                list of decks of both players.
            """
            return [deck.dict() for deck in self.decks]

        def post_deck_data(data: DeckData, str_type: Literal["str", "code"]):
            match = self.match
            deck_str = data.deck_str
            player_idx = data.player_idx
            if player_idx < 0 or player_idx > 1:
                raise HTTPException(status_code=404, detail="Player not found")
            if str_type == "str":
                deck = Deck.from_str(deck_str)
            else:
                try:
                    deck = Deck.from_deck_code(deck_str)
                except Exception:
                    raise HTTPException(status_code=403, detail="Invalid deck code")
            deck_check_res, deck_check_info = deck.check_legal(
                match.config.card_number,
                match.config.max_same_card_number,
                match.config.character_number,
                match.config.check_deck_restriction,
            )
            if not deck_check_res:
                raise HTTPException(
                    status_code=403, detail=f"Deck not legal. {deck_check_info}"
                )
            self.decks[player_idx] = deck
            try:
                deck_code = deck.to_deck_code()
                self.uploaded_deck_codes.append(deck_code)
            except Exception:
                pass

        @app.post("/deck")
        async def post_deck(data: DeckData):
            """
            Set deck.

            Args:
                data (DeckData): deck data, which contains deck_str and player_idx.
                    Note that deck_str is deck string.
            """
            post_deck_data(data, "str")

        @app.get("/deck_code/{player_idx}")
        async def get_deck_code(player_idx: int):
            """
            get deck code of a player.

            Args:
                player_idx (int): index of the player, 0 or 1.

            Return:
                deck_code: deck code of the player, which is a string.
            """
            if player_idx < 0 or player_idx > 1:
                raise HTTPException(status_code=404, detail="Player not found")
            try:
                return self.decks[player_idx].to_deck_code()
            except ValueError as e:
                raise HTTPException(status_code=403, detail=str(e))

        @app.post("/deck_code")
        async def post_deck_code(data: DeckData):
            """
            Set deck by deck code.

            Args:
                data (DeckData): deck data, which contains deck_str and player_idx.
                    Note that deck_str is deck code.
            """
            post_deck_data(data, "code")

        @app.get("/deck_code_data")
        async def get_deck_code_data():
            """
            get deck code data that is needed for generating deckcode in
            frontend.
            """
            return JSONResponse(deck_code_data)

        @app.get("/state/{mode}/{state_idx}/{player_idx}")
        async def get_game_state(
            mode: Literal["one", "after"],
            state_idx: int,
            player_idx: int,
            uuid: str | None = None,
        ):
            """
            Return list of state and its index.

            Args:
                mode (Literal["one", "after"]): mode of fetching state.
                    If mode is one, only get the state of state_idx (but also wrapped
                    in a list)
                    If mode is after: get the state of state_idx and all states after
                    it.

                state_idx (int): -1 means the last state; otherwise, it is the index of
                    the state

                player_idx (int): -1 means fetch complete data; 0 or 1 means fetch data
                    for player idx (currently not implemented)

            Return:
                list of states. For detail of state dict, refer to `post_reset`.
            """
            if state_idx > 0 and uuid != self.uuid:
                # not initial state, but uuid different
                raise HTTPException(status_code=404, detail="UUID not match")
            match = self.match
            # player idx check
            if player_idx < -1 or player_idx > 1:
                raise HTTPException(status_code=404, detail="Player not found")
            if player_idx != -1:
                raise HTTPException(
                    status_code=404, detail="player data fetch not supported"
                )
            # state idx check
            if state_idx < -1 or state_idx > len(match._history_diff):
                raise HTTPException(status_code=404, detail="State not found")
            if state_idx == -1:
                state_idx = len(match._history_diff) - 1
            start_time = time.time()
            while time.time() - start_time < self.get_state_timeout:
                if state_idx > 0 and self.uuid != uuid:
                    # during waiting data change, uuid changed. new match
                    # started, return empty result
                    return JSONResponse([])
                if state_idx == len(match._history_diff):
                    # ask for the state after the last state, wait until
                    # timeout to try to get new state
                    await asyncio.sleep(self.get_state_sleeptime)
                    continue
                result = []
                if state_idx == 0:
                    # first is 0, add full state and change state_idx to 1
                    result = [
                        {
                            "uuid": self.uuid,
                            "idx": state_idx,
                            "match": match._history[0].dict(),
                            "type": "FULL",
                        }
                    ]
                    state_idx = 1
                if mode == "after":
                    result += [
                        {
                            "uuid": self.uuid,
                            "idx": state_idx + idx,
                            "match_diff": diff,
                            "type": "DIFF",
                        }
                        for idx, diff in enumerate(match._history_diff[state_idx:])
                    ]
                return JSONResponse(result)
            return JSONResponse([])

        @app.get("/request/{player_idx}")
        async def get_request(player_idx: int):
            """
            Return the requests of player_idx. if player_idx is -1, return all
            requests. Sometimes We will update requests, but should not generate a new
            state, for example when choosing who goes first at startup. With this API,
            we can get the newest requests even if the last state is not changed.
            """
            if player_idx < -1 or player_idx > 1:
                raise HTTPException(status_code=404, detail="Player not found")
            res = self.match.requests
            if player_idx != -1:
                res = [x for x in res if x.player_idx == player_idx]
            return JSONResponse([x.dict() for x in res])

        @app.post("/respond")
        async def post_respond(data: RespondData):
            """
            Receives respond to the request. Respond command string will be sent to
            `InteractionAgent`.
            """
            if data.uuid != self.uuid:
                raise HTTPException(status_code=404, detail="UUID not match")
            if data.frame_number != -1:
                # contains frame number, should match with current frame number
                if data.frame_number + 1 < len(self.match._history_diff):
                    # old respond, ignore
                    return JSONResponse([])
            match = self.match
            player_idx = data.player_idx
            command = data.command
            current_history_length = len(match._history_diff)
            if not match.need_respond(player_idx):
                raise HTTPException(status_code=404, detail="Not your turn")
            if player_idx == 0:
                agent = self.agent_0
            elif player_idx == 1:
                agent = self.agent_1
            else:
                raise HTTPException(status_code=404, detail="Player not found")
            agent.commands = [command]
            resp = agent.generate_response(match)
            if resp is None:
                raise HTTPException(status_code=404, detail="Invalid command")
            match.respond(resp)
            match.step()
            for agent in (self.agent_0, self.agent_1):
                if (
                    agent.__class__ != InteractionAgent or len(agent.commands) > 0  # type: ignore  # noqa: E501
                ):
                    while match.need_respond(agent.player_idx):
                        resp = agent.generate_response(match)
                        assert resp is not None
                        match.respond(resp)
                        match.step()
            # after success respond, save command into command history
            self.command_history[player_idx].append(
                [
                    current_history_length - 1,
                    command,
                    len(self.command_history[0] + self.command_history[1]),
                ]
            )
            # generate response
            ret = []
            for idx, diff in enumerate(match._history_diff[current_history_length:]):
                if idx == 0 and current_history_length == 0:
                    ret.append(
                        {
                            "uuid": self.uuid,
                            "idx": idx + current_history_length,
                            "match": match._history[0].dict(),
                            "type": "FULL",
                        }
                    )
                else:
                    ret.append(
                        {
                            "uuid": self.uuid,
                            "idx": idx + current_history_length,
                            "match_diff": diff,
                            "type": "DIFF",
                        }
                    )
            return JSONResponse(ret)

        @app.get("/log")
        def get_log():
            """
            Return the log of the match. It is encoded in json.
            """
            try:
                log = self.save_log()
                return json.loads(log)
            except Exception as e:
                raise HTTPException(status_code=500, detail=str(e))

    def run(self, *argv, **kwargs):
        """
        A wrapper of uvicorn.run
        """
        if len(argv):
            raise ValueError("positional arguments not supported")
        uvicorn.run(self.app, **kwargs)

    @property
    def reset_log_save_file(self) -> str | None:
        """
        Return the save file path of reset log.
        """
        # save log
        if self.reset_log_save_path is None:
            return None
        if not os.path.exists(self.reset_log_save_path):
            os.makedirs(self.reset_log_save_path)
        filename = datetime.datetime.now().strftime("%Y%m%d_%H%M%S.%f")
        filename += ".json"
        return os.path.join(self.reset_log_save_path, filename)

    def save_log(self, file_path: str | None = None) -> str:
        """
        Save the log to specified file and return log str.
        If file path not specified, only return log str.
        log str is encoded in json.

        log format:
            match_random_state: initial random state of match
            start_deck: start deck string of players
            command_history: command history of players.
        """
        if self.match_random_state is None:
            raise RuntimeError(
                "Cannot save log when match random state is "
                "None. This may caused by reset match by "
                "match_state."
            )
        data = {
            "match_random_state": self.match_random_state,
            "start_deck": [x.to_str() for x in self.start_deck],
            "command_history": self.command_history,
            "match_config": self.match.config.dict(),
        }
        data_str = json.dumps(data)
        if file_path is not None:
            with open(file_path, "w") as f:
                f.write(data_str)
        return data_str



from typing import Any, Dict, List, Literal

from ....action import (
    ActionTypes,
    ChangeObjectUsageAction,
    MakeDamageAction,
    RemoveObjectAction,
)
from ....event import RemoveObjectEventArguments, RoundEndEventArguments
from ....match import Match
from ....status.character_status.base import ReviveCharacterStatus, UsageCharacterStatus
from .....utils.class_registry import register_class
from .....utils.desc_registry import DescDictType
from ....summon.base import AttackerSummonBase
from ....action import Actions, CreateObjectAction
from ....struct import Cost, ObjectPosition
from ....consts import (
    CostLabels,
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    IconType,
    ObjectPositionType,
    WeaponType,
)
from ....character.character_base import (
    AOESkillBase,
    CreateStatusPassiveSkill,
    ElementalBurstBase,
    ElementalNormalAttackBase,
    ElementalSkillBase,
    CharacterBase,
    TalentBase,
)


class OverwhelmingIce_4_4(UsageCharacterStatus):
    name: Literal["Overwhelming Ice"] = "Overwhelming Ice"
    version: Literal["4.4"] = "4.4"
    usage: int = 1
    max_usage: int = 1
    icon_type: IconType = IconType.ATK_SELF


class CryocrystalCore_4_4(ReviveCharacterStatus):
    name: Literal["Cryocrystal Core"] = "Cryocrystal Core"
    version: Literal["4.4"] = "4.4"
    heal: int = 1


class PiercingIceridge_4_4(AttackerSummonBase):
    name: Literal["Piercing Iceridge"] = "Piercing Iceridge"
    version: Literal["4.4"] = "4.4"
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.CRYO
    damage: int = 1

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Match
    ) -> List[MakeDamageAction | ChangeObjectUsageAction]:
        ret = super().event_handler_ROUND_END(event, match)
        damage_action = ret[0]
        assert damage_action.type == ActionTypes.MAKE_DAMAGE
        # get closest character
        our_active = match.player_tables[self.position.player_idx].active_character_idx
        target_characters = match.player_tables[1 - self.position.player_idx].characters
        target_idx = our_active
        if target_characters[target_idx].is_defeated:
            # need to choose another character
            for idx, character in enumerate(target_characters):
                if not character.is_defeated:
                    target_idx = idx
                    break
            else:
                raise AssertionError("No character alive")
        damage_action.damage_value_list[0].target_position = target_characters[
            target_idx
        ].position
        return ret


# Skills


class IcespikeShot(ElementalNormalAttackBase, AOESkillBase):
    name: Literal["Icespike Shot"] = "Icespike Shot"
    version: Literal["4.4"] = "4.4"
    damage: int = 1
    back_damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.CRYO, elemental_dice_number=1, any_dice_number=2
    )

    def get_actions(self, match: Match) -> List[Actions]:
        found_target = self.query_one(match, "self status 'name=Overwhelming Ice'")
        if found_target is not None:
            # use AOE attack opposite
            return [
                self.attack_opposite_active(match, self.damage, self.damage_type),
                self.charge_self(1),
                RemoveObjectAction(object_position=found_target.position),
            ]
        # use normal attack opposite
        return [
            ElementalNormalAttackBase.attack_opposite_active(
                self, match, self.damage, self.damage_type
            ),
            self.charge_self(1),
        ]


class IceRingWaltz(ElementalSkillBase):
    name: Literal["Ice Ring Waltz"] = "Ice Ring Waltz"
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(elemental_dice_color=DieColor.CRYO, elemental_dice_number=3)

    def get_actions(self, match: Match) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [
            self.create_character_status("Overwhelming Ice"),
        ]


class PlungingIceShards(ElementalBurstBase, AOESkillBase):
    name: Literal["Plunging Ice Shards"] = "Plunging Ice Shards"
    damage: int = 2
    back_damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.CRYO,
        elemental_dice_number=3,
        charge=2,
    )

    def get_actions(self, match: Match) -> List[Actions]:
        return super().get_actions(match) + [self.create_summon("Piercing Iceridge")]


class CryocrystalCore(CreateStatusPassiveSkill):
    name: Literal["Cryocrystal Core"] = "Cryocrystal Core"
    status_name: Literal["Cryocrystal Core"] = "Cryocrystal Core"
    regenerate_when_revive: bool = False


# Talents


class SternfrostPrism_4_4(TalentBase):
    name: Literal["Sternfrost Prism"]
    version: Literal["4.4"] = "4.4"
    character_name: Literal["Cryo Hypostasis"] = "Cryo Hypostasis"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.CRYO,
        elemental_dice_number=1,
    )
    cost_label: int = (
        CostLabels.CARD.value | CostLabels.TALENT.value | CostLabels.EQUIPMENT.value
    )
    remove_when_used: bool = False

    def get_actions(self, target: ObjectPosition | None, match: Match) -> List[Actions]:
        """
        Using this card will attach cryo crystal core to it.
        """
        ret = super().get_actions(target, match)
        character: Any = self.query_one(match, "our active")
        assert character is not None and character.name == self.character_name
        return ret + [
            CreateObjectAction(
                object_name="Cryocrystal Core",
                object_position=character.position.set_area(
                    ObjectPositionType.CHARACTER_STATUS
                ),
                object_arguments={},
            )
        ]

    def event_handler_REMOVE_OBJECT(
        self, event: RemoveObjectEventArguments, match: Match
    ) -> List[CreateObjectAction]:
        """
        When this card equipped, cryo core is removed from self, create
        sheer cold on opposite active.
        """
        if self.position.not_satisfy(
            "both pidx=same cidx=same "
            "and source area=character "
            "and target area=character_status",
            event.action.object_position,
        ):
            # not equipped, or target not self character status
            return []
        if event.object_name != "Cryocrystal Core":
            # not cryo crystal core
            return []
        target = match.player_tables[
            1 - self.position.player_idx
        ].get_active_character()
        return [
            CreateObjectAction(
                object_position=target.position.set_area(
                    ObjectPositionType.CHARACTER_STATUS
                ),
                object_name="Sheer Cold",
                object_arguments={},
            )
        ]


class CryoHypostasis_4_4(CharacterBase):
    name: Literal["Cryo Hypostasis"]
    version: Literal["4.4"] = "4.4"
    element: ElementType = ElementType.CRYO
    max_hp: int = 8
    max_charge: int = 2
    skills: List[IcespikeShot | IceRingWaltz | PlungingIceShards | CryocrystalCore] = []
    faction: List[FactionType] = [FactionType.MONSTER]
    weapon_type: WeaponType = WeaponType.OTHER

    def _init_skills(self) -> None:
        self.skills = [
            IcespikeShot(),
            IceRingWaltz(),
            PlungingIceShards(),
            CryocrystalCore(),
        ]


# define descriptions of newly defined classes. Note key of skills and talents
# have character names. For balance changes, only descs are needed to define.
character_descs: Dict[str, DescDictType] = {
    "CHARACTER/Cryo Hypostasis": {
        "names": {"en-US": "Cryo Hypostasis", "zh-CN": "无相之冰"},
        "descs": {"4.4": {"en-US": "", "zh-CN": ""}},
        "image_path": "cardface/Char_Monster_EffigyIce.png",  # noqa: E501
        "id": 2103,
    },
    "SKILL_Cryo Hypostasis_NORMAL_ATTACK/Icespike Shot": {
        "names": {"en-US": "Icespike Shot", "zh-CN": "冰锥迸射"},
        "descs": {
            "4.4": {"en-US": "Deals 1 Cryo DMG.", "zh-CN": "造成1点冰元素伤害。"}
        },
    },
    "SKILL_Cryo Hypostasis_ELEMENTAL_SKILL/Ice Ring Waltz": {
        "names": {"en-US": "Ice Ring Waltz", "zh-CN": "圆舞冰环"},
        "descs": {
            "4.4": {
                "en-US": "Deals 3 Cryo DMG. This character gains Overwhelming Ice.",  # noqa: E501
                "zh-CN": "造成3点冰元素伤害，本角色附属四迸冰锥。",
            }
        },
    },
    "CHARACTER_STATUS/Overwhelming Ice": {
        "names": {"en-US": "Overwhelming Ice", "zh-CN": "四迸冰锥"},
        "descs": {
            "4.4": {
                "en-US": "When the character uses a Normal Attack: Deal 1 Piercing DMG to all opposing characters on standby. .\nUsage(s): 1",  # noqa: E501
                "zh-CN": "角色进行普通攻击时：对所有敌方后台角色造成1点穿透伤害。\n可用次数：1",  # noqa: E501
            }
        },
    },
    "SKILL_Cryo Hypostasis_ELEMENTAL_BURST/Plunging Ice Shards": {
        "names": {"en-US": "Plunging Ice Shards", "zh-CN": "冰棱轰坠"},
        "descs": {
            "4.4": {
                "en-US": "Deals 2 Cryo DMG, deals 1 Piercing DMG to all opposing characters on standby, summons 1 Piercing Iceridge.",  # noqa: E501
                "zh-CN": "造成2点冰元素伤害，对所有敌方后台角色造成1点穿透伤害，召唤刺击冰棱。",  # noqa: E501
            }
        },
    },
    "SUMMON/Piercing Iceridge": {
        "names": {"en-US": "Piercing Iceridge", "zh-CN": "刺击冰棱"},
        "descs": {
            "4.4": {
                "en-US": "End Phase: Deal 1 Cryo DMG to the opposing Character Closest to Your Current Active Character.\nUsage(s): 2",  # noqa: E501
                "zh-CN": "结束阶段：对敌方距离我方出战角色最近的角色造成1点冰元素伤害。\n可用次数：2",  # noqa: E501
            }
        },
        "image_path": "cardface/Summon_EffigyIce.png",
    },
    "SKILL_Cryo Hypostasis_PASSIVE/Cryocrystal Core": {
        "names": {"en-US": "Cryocrystal Core", "zh-CN": "冰晶核心"},
        "descs": {
            "4.4": {
                "en-US": "(Passive) When the battle begins, this character gains Cryocrystal Core.",  # noqa: E501
                "zh-CN": "【被动】战斗开始时，初始附属冰晶核心。",
            }
        },
    },
    "CHARACTER_STATUS/Cryocrystal Core": {
        "names": {"en-US": "Cryocrystal Core", "zh-CN": "冰晶核心"},
        "descs": {
            "4.4": {
                "en-US": "When the character to which this is attached would be defeated: Remove this effect, ensure the character $[K54|s1], and heal them to 1 HP.",  # noqa: E501
                "zh-CN": "所附属角色被击倒时：移除此效果，使角色免于被击倒，并治疗该角色到1点生命值。",  # noqa: E501
            }
        },
    },
    "TALENT_Cryo Hypostasis/Sternfrost Prism": {
        "names": {"en-US": "Sternfrost Prism", "zh-CN": "严霜棱晶"},
        "descs": {
            "4.4": {
                "en-US": "Can only be played if your active character is Cryo Hypostasis: Attach Cryocrystal Core to them.\nAfter Cryo Hypostasis, who has this card equipped, triggers Cryocrystal Core: Attach Sheer Cold to the opposing active character.\n(Your deck must contain Cryo Hypostasis to add this card to your deck)",  # noqa: E501
                "zh-CN": "我方出战角色为无相之冰时，才能打出：使其附属冰晶核心。\n装备有此牌的无相之冰触发冰晶核心后：对敌方出战角色附属严寒。\n（牌组中包含无相之冰，才能加入牌组）",  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Talent_EffigyIce.png",  # noqa: E501
        "id": 221031,
    },
}


register_class(
    CryoHypostasis_4_4
    | OverwhelmingIce_4_4
    | CryocrystalCore_4_4
    | PiercingIceridge_4_4
    | SternfrostPrism_4_4,
    character_descs,
)



from lpsim.server.struct import ObjectPosition
from lpsim.agents.interaction_agent import InteractionAgent
from lpsim.server.match import Match, MatchState
from lpsim.server.deck import Deck
from tests.utils_for_test import (
    check_hp,
    get_random_state,
    get_test_id_from_command,
    make_respond,
    set_16_omni,
)


def test_adeptus_temptation():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "TEST 1 all card can use 3 target",
            "card 0 0 0 1",
            "skill 0 0 1 2",
            "skill 1 0 1 2",
            "skill 1 0 1 2",
            "TEST 2 10 10 10 10 10 6",
            "skill 2 0 1 2",
            "TEST 3 status 1 sati, 8 8 0, card target 1 2",
            "end",
            "TEST 4 p0c0 no status, 3 can eat",
            "end",
            "end",
            "sw_char 2 0",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "skill 3 0 1 2",
            "TEST 2 1 8 8 6 2 0",
            "end",
        ],
        [
            "sw_card",
            "choose 2",
            "end",
            "choose 0",
            "card 0 0 0 1",
            "card 0 0 0 1",
            "TEST 5 cannot eat",
            "skill 1 0 1 2",
            "skill 1 0 1 2",
            "skill 1 0 1 2",
            "end",
            "skill 2 0 1 2",
            "TEST 2 2 8 8 6 8 0",
            "end",
            "sw_char 1 0",
            "card 0 1 0 1",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    match = Match(random_state=get_random_state())
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Fischl
        character:Mona
        character:Nahida
        # Gambler's Earrings*2
        # Wine-Stained Tricorne*2
        # Vanarana
        # Timmie*2
        # Rana*2
        # Covenant of Rock
        # Wind and Freedom
        # The Bestest Travel Companion!*2
        # Changing Shifts*2
        # Toss-Up
        # Strategize*2
        # I Haven't Lost Yet!*2
        # Leave It to Me!
        # Calx's Arts*2
        # Adeptus' Temptation*2
        Adeptus' Temptation*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    match.config.random_first_player = False
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 1:
                count = 0
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        count += 1
                        assert len(req.targets) == 3
                assert count > 0
            elif test_id == 2:
                hps = [int(x) for x in cmd.strip().split()[2:]]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 3:
                check_hp(match, [[10, 10, 10], [8, 8, 0]])
                table = match.player_tables[0]
                assert len(table.characters[0].status) == 1
                assert table.characters[0].status[0].name == "Satiated"
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        assert len(req.targets) == 2
                        assert isinstance(req.targets[0], ObjectPosition)
                        assert isinstance(req.targets[1], ObjectPosition)
                        assert req.targets[0].character_idx == 1
                        assert req.targets[1].character_idx == 2
            elif test_id == 4:
                assert len(match.player_tables[0].characters[0].status) == 0
                cards = 0
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        assert len(req.targets) == 3
                        cards += 1
                assert cards > 0
            elif test_id == 5:
                for req in match.requests:
                    assert req.name != "UseCardRequest"
            else:
                break
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    assert match.state != MatchState.ERROR


def test_lotus_flower_crisp():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "TEST 1 card can use",
            "card 0 0 1",
            "skill 1 0 1 2",
            "TEST 2 10 10 10 9 10 10",
            "TEST 3 status 1",
            "end",
            "TEST 2 9 10 10 9 10 10",
            "sw_char 1 0",
            "sw_char 0 0",
            "skill 1 0 1 2",
            "sw_char 1 0",
            "card 0 1 0",
            "TEST 4 card target 0 2",
            "skill 1 0 1 2",
            "TEST 2 5 6 10 8 9 7",
            "sw_char 0 0",
            "end",
        ],
        [
            "sw_card",
            "choose 0",
            "TEST 5 p0c0 2 status",
            "skill 1 0 1 2",
            "card 0 0 0",
            "end",
            "sw_char 2 0",
            "skill 1 0 1 2",
            "skill 1 0 1 2",
            "skill 3 0 1 2",
            "card 0 0 0",
            "card 0 0 0",
            "card 0 0 0",
            "TEST 6 no card use",
            "end",
            "TEST 2 4 6 10 7 8 5",
            "TEST 7 p1 no c status",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    match = Match(random_state=get_random_state())
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Fischl
        character:Mona
        character:Nahida
        # Gambler's Earrings*2
        # Wine-Stained Tricorne*2
        # Vanarana
        # Timmie*2
        # Rana*2
        # Covenant of Rock
        # Wind and Freedom
        # The Bestest Travel Companion!*2
        # Changing Shifts*2
        # Toss-Up
        # Strategize*2
        # I Haven't Lost Yet!*2
        # Leave It to Me!
        # Calx's Arts*2
        # Adeptus' Temptation*2
        # Lotus Flower Crisp*2
        Lotus Flower Crisp*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    match.config.random_first_player = False
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 1:
                count = 0
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        count += 1
                        assert len(req.targets) == 3
                assert count > 0
            elif test_id == 2:
                hps = [int(x) for x in cmd.strip().split()[2:]]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 3:
                table = match.player_tables[0]
                assert len(table.characters[0].status) == 1
                assert table.characters[0].status[0].name == "Satiated"
            elif test_id == 4:
                cards = 0
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        assert len(req.targets) == 2
                        assert isinstance(req.targets[0], ObjectPosition)
                        assert isinstance(req.targets[1], ObjectPosition)
                        assert req.targets[0].character_idx == 0
                        assert req.targets[1].character_idx == 2
                        cards += 1
                assert cards > 0
            elif test_id == 5:
                table = match.player_tables[0]
                assert len(table.characters[0].status) == 2
            elif test_id == 6:
                for req in match.requests:
                    assert req.name != "UseCardRequest"
            elif test_id == 7:
                table = match.player_tables[1]
                for character in table.characters:
                    assert len(character.status) == 0
            else:
                break
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    assert match.state != MatchState.ERROR


def test_lotus_flower_crisp_and_reflection():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "TEST 1 card can use",
            "card 0 0 1",
            "skill 1 0 1 2",
            "TEST 2 10 10 10 9 10 10",
            "TEST 3 status 1",
            "end",
            "TEST 2 9 10 10 9 10 10",
            "sw_char 1 0",
            "sw_char 0 0",
            "skill 1 0 1 2",
            "sw_char 1 0",
            "card 0 1 0",
            "TEST 4 card target 0 2",
            "skill 1 0 1 2",
            "sw_char 0 0",
        ],
        [
            "sw_card",
            "choose 0",
            "TEST 5 p0c0 2 status",
            "skill 1 0 1 2",
            "card 0 0 0",
            "end",
            "sw_char 2 0",
            "skill 1 0 1 2",
            "skill 1 0 1 2",
            "skill 0 0 1 2",
            "TEST 6 5 7 10 8 9 7 and reflection remain usage",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    match = Match(random_state=get_random_state())
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Fischl
        character:Mona
        character:Nahida
        # Gambler's Earrings*2
        # Wine-Stained Tricorne*2
        # Vanarana
        # Timmie*2
        # Rana*2
        # Covenant of Rock
        # Wind and Freedom
        # The Bestest Travel Companion!*2
        # Changing Shifts*2
        # Toss-Up
        # Strategize*2
        # I Haven't Lost Yet!*2
        # Leave It to Me!
        # Calx's Arts*2
        # Adeptus' Temptation*2
        # Lotus Flower Crisp*2
        Lotus Flower Crisp*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    match.config.random_first_player = False
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 1:
                count = 0
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        count += 1
                        assert len(req.targets) == 3
                assert count > 0
            elif test_id == 2:
                hps = [int(x) for x in cmd.strip().split()[2:]]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 3:
                table = match.player_tables[0]
                assert len(table.characters[0].status) == 1
                assert table.characters[0].status[0].name == "Satiated"
            elif test_id == 4:
                cards = 0
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        assert len(req.targets) == 2
                        assert isinstance(req.targets[0], ObjectPosition)
                        assert isinstance(req.targets[1], ObjectPosition)
                        assert req.targets[0].character_idx == 0
                        assert req.targets[1].character_idx == 2
                        cards += 1
                assert cards > 0
            elif test_id == 5:
                table = match.player_tables[0]
                assert len(table.characters[0].status) == 2
            elif test_id == 6:
                # "TEST 6 5 7 10 8 9 7 and reflection remain usage",
                check_hp(match, [[5, 7, 10], [8, 9, 7]])
                table = match.player_tables[0]
                assert len(table.summons) == 2
                assert table.summons[1].name == "Reflection"
                assert table.summons[1].usage == 1
            else:
                break
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    assert match.state != MatchState.ERROR


def test_mond_hash():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "TEST 1 no card can use",
            "sw_char 2 0",
            "TEST 2 card can use one target",
            "card 0 0 0",
            "TEST 1 no use card",
            "TEST 3 10 10 10 10 10 10",
            "end",
        ],
        ["sw_card", "choose 0", "TEST 1 no card can use", "skill 1 0 1 2"],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    match = Match(random_state=get_random_state())
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Fischl
        character:Mona
        character:Nahida
        # Gambler's Earrings*2
        # Wine-Stained Tricorne*2
        # Vanarana
        # Timmie*2
        # Rana*2
        # Covenant of Rock
        # Wind and Freedom
        # The Bestest Travel Companion!*2
        # Changing Shifts*2
        # Toss-Up
        # Strategize*2
        # I Haven't Lost Yet!*2
        # Leave It to Me!
        # Calx's Arts*2
        # Adeptus' Temptation*2
        # Lotus Flower Crisp*2
        # Mondstadt Hash Brown*2
        Mondstadt Hash Brown*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    match.config.random_first_player = False
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 1:
                for req in match.requests:
                    assert req.name != "UseCardRequest"
            elif test_id == 2:
                count = 0
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        count += 1
                        assert len(req.targets) == 1
                assert count > 0
            elif test_id == 3:
                hps = [int(x) for x in cmd.strip().split()[2:]]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            else:
                break
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    assert match.state != MatchState.ERROR


def test_tandoori():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "TEST 1 2 card can use",
            "sw_char 2 0",
            "TEST 2 5 card can use",
            "card 0 0 0",
            "card 1 0 0 1",
            "TEST 3 p0c0c1 2 status p0c2 1 no card can use",
            "sw_char 0 0",
            "skill 0 0 1 2",
            "TEST 4 not consumed",
            "TEST 5 10 10 10 10 10 8",
            "skill 1 0 1 2",
            "TEST 5 10 10 10 7 10 8",
            "sw_char 1 0",
            "skill 1 0 1 2",
            "TEST 5 10 8 10 7 10 5",
            "end",
            "TEST 6 p0 only c1 status seed",
            "end",
        ],
        [
            "sw_card",
            "choose 0",
            "skill 1 0 1 2",
            "sw_char 2 0",
            "sw_char 0 0",
            "sw_char 2 0",
            "skill 1 0 1 2",
            "end",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    match = Match(random_state=get_random_state())
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Fischl
        character:Mona
        character:Nahida
        # Gambler's Earrings*2
        # Wine-Stained Tricorne*2
        # Vanarana
        # Timmie*2
        # Rana*2
        # Covenant of Rock
        # Wind and Freedom
        # The Bestest Travel Companion!*2
        # Changing Shifts*2
        # Toss-Up
        # Strategize*2
        # I Haven't Lost Yet!*2
        # Leave It to Me!
        # Calx's Arts*2
        # Adeptus' Temptation*2
        # Lotus Flower Crisp*2
        # Mondstadt Hash Brown*2
        # Tandoori Roast Chicken
        Tandoori Roast Chicken*15
        Mondstadt Hash Brown*15
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    match.config.random_first_player = False
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 1:
                count = 0
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        count += 1
                        assert len(req.targets) == 0
                assert count == 2
            elif test_id == 2:
                count = 0
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        count += 1
                assert count == 5
            elif test_id == 3:
                characters = match.player_tables[0].characters
                assert len(characters[0].status) == 2
                assert len(characters[1].status) == 2
                assert len(characters[2].status) == 1
                for req in match.requests:
                    assert req.name != "UseCardRequest"
            elif test_id == 4:
                characters = match.player_tables[0].characters
                assert len(characters[0].status) == 2
                assert len(characters[1].status) == 2
                assert len(characters[2].status) == 1
            elif test_id == 5:
                hps = [int(x) for x in cmd.strip().split()[2:]]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 6:
                characters = match.player_tables[0].characters
                assert len(characters[0].status) == 0
                assert len(characters[1].status) == 1
                assert characters[1].status[0].name == "Seed of Skandha"
                assert len(characters[2].status) == 0
            else:
                break
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    assert match.state != MatchState.ERROR


def test_pizza():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "TEST 2 no card can use",
            "skill 1 0 1 2",
            "end",
            "TEST 3 p1c0 status 2 1",
            "skill 1 0 1 2",
            "end",
            "end",
        ],
        [
            "sw_card",
            "choose 0",
            "TEST 4 all card can use",
            "card 0 0 0",
            "end",
            "TEST 3 p1c0 status 1",
            "card 0 0 0",
            "end",
            "end",
            "TEST 1 10 10 10 9 10 10",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:ElectroMobMage
        character:CryoMobMage
        character:Noelle
        Mushroom Pizza*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                for req in match.requests:
                    assert req.name != "UseCardRequest"
            elif test_id == 3:
                cmd = cmd.split()
                status = match.player_tables[1].characters[0].status
                usages = [int(x) for x in cmd[4:]]
                assert len(usages) == len(status)
                for s, u in zip(status, usages):
                    assert s.usage == u
            elif test_id == 4:
                count = 0
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        count += 1
                assert count == 5
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_full_and_tandoori():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "skill 0 0 1 2",
            "card 2 0 0",
            "card 0 0 0 1",
            "skill 1 0 1 2",
            "end",
            "card 0 0 0 1",
            "skill 1 0 1 2",
            "sw_char 1 0",
            "skill 1 0 1 2",
            "TEST 1 10 10 10 7 7 10",
            "end",
            "end",
        ],
        [
            "sw_card",
            "choose 0",
            "skill 0 0 1 2",
            "TEST 1 9 10 10 7 10 10",
            "end",
            "sw_char 1 0",
            "card 0 0 0",
            "TEST 1 10 10 10 7 7 10",
            "end",
            "card 7 0 0 1",
            "skill 0 0 1 2",
            "sw_char 0 0",
            "skill 1 0 1 2",
            "TEST 1 9 7 9 8 6 10",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Fischl
        character:Rhodeia of Loch
        character:Nahida
        Mushroom Pizza*15
        Tandoori Roast Chicken*15
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_north_chicken():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "TEST 1 card can use",
            "TEST 2 skill 0 cost 3",
            "card 0 0",
            "TEST 2 skill 0 cost 2",
            "TEST 2 skill 1 cost 3",
            "sw_char 1 0",
            "TEST 2 skill 0 cost 3",
            "sw_char 0 0",
            "skill 0 0 1",
            "TEST 2 skill 0 cost 3",
            "TEST 3 p0c0 status 1",
            "end",
            "card 0 0",
            "end",
        ],
        [
            "sw_card",
            "choose 0",
            "end",
            "sw_char 1 0",
            "sw_char 0 0",
            "TEST 2 skill 0 cost 3",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Arataki Itto
        character:Mona
        character:Nahida
        Northern Smoked Chicken*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                count = 0
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        count += 1
                assert count > 0
            elif test_id == 2:
                cmd = cmd.split()
                sidx = int(cmd[3])
                c = int(cmd[5])
                for req in match.requests:
                    if req.name == "UseSkillRequest":
                        if req.skill_idx == sidx:
                            assert req.cost.total_dice_cost == c
            elif test_id == 3:
                assert len(match.player_tables[0].characters[0].status) == 1
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_north_chicken_2():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "skill 1 0 1 2",
            "skill 1 0 1 2",
            "end",
            "card 0 0 0 1",
            "card 1 0",
            "skill 0 1",
            "sw_char 1 0",
            "sw_char 2 0",
            "sw_char 0 0",
            "TEST 1 p0c0 status 3",
            "skill 0 1",
            "end",
        ],
        ["sw_card", "choose 0", "skill 1 0 1 2", "end", "end"],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Arataki Itto
        character:Mona
        character:Nahida
        Northern Smoked Chicken*15
        Laurel Coronet*15
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                assert len(match.player_tables[0].characters[0].status) == 3
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_guoba_meatroll_egg_sashimi_crab():
    cmd_records = [
        [
            "sw_card 1 2",
            "choose 0",
            "card 0 0 15",
            "skill 2 14 13 12 11 10",
            "TEST 1 9 10 10 8 7 8",
            "sw_char 1 9",
            "TEST 1 9 7 10 8 7 8",
            "TEST 2 p1c1 usage 1",
            "skill 1 8 7 6",
            "sw_char 0 5",
            "end",
            "card 4 0 15",
            "skill 2 14 13 12 11 10",
            "TEST 1 8 7 10 4 0 4",
            "skill 0 9 8 7",
            "TEST 1 8 7 10 1 0 4",
            "TEST 2 p0c0 usage 1 1",
            "TEST 3 card 2 cannot use",
            "end",
            "card 0 0 15 0",
            "sw_char 1 1",
            "TEST 1 5 6 10 0 0 4",
            "sw_char 2 1",
            "sw_char 1 1",
            "TEST 1 2 2 7 0 0 4",
            "sw_char 0 1",
            "sw_char 1 1",
            "choose 2",
            "TEST 1 2 0 7 0 0 4",
            "TEST 3 card 0 cannot use",
            "card 1 0 1 8 7",
            "TEST 4 no card can use",
            "end",
            "sw_char 1 15",
            "choose 0",
            "end",
            "choose 2",
            "card 1 0 15 14 13",
            "TEST 3 card 6 cannot use",
            "sw_char 0 12",
            "choose 2",
            "end",
        ],
        [
            "sw_card 4 2",
            "choose 1",
            "TEST 1 10 10 10 8 7 8",
            "card 1 1",
            "skill 1 15 14 13",
            "skill 0 12 11 10",
            "end",
            "end",
            "choose 0",
            "choose 2",
            "card 5 0 15",
            "skill 0 14 13",
            "skill 1 12 11 10",
            "skill 0 9 8",
            "skill 2 7 6 5 4",
            "skill 0 3 2",
            "skill 0 1 0",
            "end",
            "card 5 0",
            "skill 1 15 14 13",
            "skill 0 12 11 10",
            "end",
            "card 6 0 15",
            "skill 0 14 13",
            "skill 0 12 11",
            "skill 0 10 9",
            "TEST 5 skill 0 cost 3",
            "TEST 2 p1c2 usage 1 1",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Ganyu
        character:Fischl
        character:Keqing
        Jueyun Guoba*5
        Minty Meat Rolls*5
        Minty Meat Rolls@3.3*5
        Teyvat Fried Egg*5
        Sashimi Platter*5
        Butter Crab*5
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                cidx = int(cmd[2][3])
                usage = [int(x) for x in cmd[4:]]
                status = match.player_tables[pidx].characters[cidx].status
                assert len(status) == len(usage)
                for s, u in zip(status, usage):
                    assert s.usage == u
            elif test_id == 3:
                cmd = cmd.split()
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        assert req.card_idx != int(cmd[3])
            elif test_id == 4:
                for req in match.requests:
                    assert req.name != "UseCardRequest"
            elif test_id == 5:
                for req in match.requests:
                    if req.name == "UseSkillRequest" and req.skill_idx == 0:
                        assert req.cost.total_dice_cost == 3
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_meatroll():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "skill 1 0 1 2",
            "skill 1 0 1 2",
            "end",
            "card 0 0 0 1",
            "card 1 0 1",
            "card 2 0 1",
            "skill 0 1",
            "sw_char 1 0",
            "sw_char 2 0",
            "sw_char 0 0",
            "TEST 1 p0c0 status 3",
            "skill 0 1",
            "end",
        ],
        ["sw_card", "choose 0", "skill 1 0 1 2", "end", "end"],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Arataki Itto
        character:Mona
        character:Nahida
        Minty Meat Rolls*15
        Laurel Coronet*15
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                assert len(match.player_tables[0].characters[0].status) == 3
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_guoba_egg():
    cmd_records = [
        [
            "sw_card 1 2",
            "choose 0",
            "card 0 0 15",
            "skill 1 14 13 12",
            "TEST 1 8 8 8 9 10 10",
            "TEST 2 card 2 cannot use",
            "sw_char 2 11",
            "sw_char 0 10",
            "skill 0 9 8 7",
            "skill 0 6 5 4",
            "skill 1 3 2 1",
            "end",
            "choose 2",
            "card 2 0 15 14 13",
            "sw_char 0 12",
            "skill 1 11 10 9",
            "skill 0 8 7 6",
            "choose 1",
            "TEST 2 card 2 cannot use",
            "TEST 1 0 7 7 9 9 1",
            "end",
        ],
        [
            "sw_card 2 3 4 0",
            "choose 0",
            "TEST 1 10 10 10 9 10 10",
            "card 0 0",
            "skill 2 15 14 13 12 11",
            "sw_char 2 10",
            "skill 1 9 8 7",
            "skill 1 6 5 4",
            "end",
            "skill 1 15 14 13",
            "sw_char 1 12",
            "sw_char 2 11",
            "skill 0 10 9 8",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Ganyu
        character:Fischl
        character:Keqing
        Jueyun Guoba*5
        Minty Meat Rolls*5
        Minty Meat Rolls@3.3*5
        Teyvat Fried Egg*5
        Sashimi Platter*5
        Butter Crab*5
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                cmd = cmd.split()
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        assert req.card_idx != int(cmd[3])
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == "__main__":
    # test_adeptus_temptation()
    # test_lotus_flower_crisp()
    # test_lotus_flower_crisp_and_reflection()
    # test_mond_hash()
    test_tandoori()
    # test_pizza()
    # test_full_and_tandoori()
    # test_north_chicken()
    # test_north_chicken_2()
    test_guoba_meatroll_egg_sashimi_crab()
    # test_meatroll()
    test_guoba_egg()



from typing import Any, Literal

from .....utils.class_registry import register_class

from ....consts import ObjectPositionType, SkillType, WeaponType
from ....modifiable_values import DamageIncreaseValue
from ....struct import Cost
from .base import RoundEffectWeaponBase


class SkywardBase(RoundEffectWeaponBase):
    name: str
    cost: Cost = Cost(same_dice_number=3)
    version: str
    weapon_type: WeaponType
    max_usage_per_round: int = 1

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        if self.position.area != ObjectPositionType.CHARACTER:
            # not equipped
            return value
        super().value_modifier_DAMAGE_INCREASE(value, match, mode)
        if (
            value.is_corresponding_character_use_damage_skill(
                self.position, match, SkillType.NORMAL_ATTACK
            )
            and self.usage > 0
        ):
            # have usage and is normal attack
            assert mode == "REAL"
            self.usage -= 1
            value.damage += 1
        return value


class SkywardAtlas_3_3(SkywardBase):
    name: Literal["Skyward Atlas"]
    version: Literal["3.3"] = "3.3"
    weapon_type: Literal[WeaponType.CATALYST] = WeaponType.CATALYST


class SkywardHarp_3_3(SkywardBase):
    name: Literal["Skyward Harp"]
    version: Literal["3.3"] = "3.3"
    weapon_type: Literal[WeaponType.BOW] = WeaponType.BOW


class SkywardSpine_3_3(SkywardBase):
    name: Literal["Skyward Spine"]
    version: Literal["3.3"] = "3.3"
    weapon_type: Literal[WeaponType.POLEARM] = WeaponType.POLEARM


class SkywardBlade_3_7(SkywardBase):
    name: Literal["Skyward Blade"]
    version: Literal["3.7"] = "3.7"
    weapon_type: Literal[WeaponType.SWORD] = WeaponType.SWORD


class SkywardPride_3_7(SkywardBase):
    name: Literal["Skyward Pride"]
    version: Literal["3.7"] = "3.7"
    weapon_type: Literal[WeaponType.CLAYMORE] = WeaponType.CLAYMORE


register_class(
    SkywardAtlas_3_3
    | SkywardHarp_3_3
    | SkywardPride_3_7
    | SkywardSpine_3_3
    | SkywardBlade_3_7
)



from typing import Any, List, Literal

from ...action import Actions

from ....utils.class_registry import register_class

from ..character_base import ElementalSkillBase, PhysicalNormalAttackBase
from ...struct import Cost
from ...consts import DamageElementalType, DieColor
from .kamisato_ayato_4_1 import KamisatoAyato_4_1 as KA_4_1
from .kamisato_ayato_4_1 import KamisatoArtSuiyuu as KAS_4_1


class KamisatoArtSuiyuu(KAS_4_1):
    damage: int = 3
    cost: Cost = Cost(
        elemental_dice_color=DieColor.HYDRO, elemental_dice_number=3, charge=3
    )


class KamisatoArtKyouka(ElementalSkillBase):
    name: Literal["Kamisato Art: Kyouka"] = "Kamisato Art: Kyouka"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(elemental_dice_color=DieColor.HYDRO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [
            self.create_character_status("Takimeguri Kanka"),
        ]


class KamisatoAyato_3_6(KA_4_1):
    version: Literal["3.6"]
    max_charge: int = 3
    skills: List[PhysicalNormalAttackBase | KamisatoArtKyouka | KamisatoArtSuiyuu] = []

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Kamisato Art: Marobashi",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            KamisatoArtKyouka(),
            KamisatoArtSuiyuu(),
        ]


register_class(KamisatoAyato_3_6)



import pytest
from lpsim.server.deck import Deck


def test_create_mob():
    TEST_DECK = {
        "name": "Deck",
        "characters": [
            {
                "name": "PhysicalMob",
                "element": "ELECTRO",
            },
            {
                "name": "ElectroMobMage",
                "element": "ELECTRO",
            },
            {
                "name": "ElectroMob",
                "element": "ELECTRO",
            },
            {
                "name": "DendroMob",
            },
            {
                "name": "DendroMobMage",
            },
        ],
        "cards": [
            {
                "name": "Strategize",
            }
        ]
        * 30,
    }
    Deck(**TEST_DECK)
    TEST_DECK["characters"][1]["element"] = "DENDRO"
    with pytest.raises(ValueError):
        Deck(**TEST_DECK)
    TEST_DECK["characters"][1]["element"] = "ELECTRO"
    TEST_DECK["characters"][2]["element"] = "DENDRO"
    with pytest.raises(ValueError):
        Deck(**TEST_DECK)
    TEST_DECK["characters"][2]["element"] = "ELECTRO"



from typing import Any, Literal

from .....utils.class_registry import register_class

from .base import RoundEffectArtifactBase
from ....struct import Cost
from ....modifiable_values import CostValue, InitialDiceColorValue
from ....consts import ELEMENT_TO_DIE_COLOR, ElementType, ObjectPositionType


class SmallElementalArtifact_4_0(RoundEffectArtifactBase):
    """
    Seven artifacts that decrease elemental cost.
    """

    name: Literal[
        "Broken Rime's Echo",  # cryo
        "Laurel Coronet",  # dendro
        "Mask of Solitude Basalt",  # geo
        "Thunder Summoner's Crown",  # electro
        "Viridescent Venerer's Diadem",  # anemo
        "Wine-Stained Tricorne",  # hydro
        "Witch's Scorching Hat",  # pyro
    ]

    version: Literal["4.0"] = "4.0"
    usage: int = 1
    cost: Cost = Cost(any_dice_number=2)
    element: ElementType = ElementType.NONE
    max_usage_per_round: int = 1

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        if self.name == "Broken Rime's Echo":
            self.element = ElementType.CRYO
        elif self.name == "Laurel Coronet":
            self.element = ElementType.DENDRO
        elif self.name == "Mask of Solitude Basalt":
            self.element = ElementType.GEO
        elif self.name == "Thunder Summoner's Crown":
            self.element = ElementType.ELECTRO
        elif self.name == "Viridescent Venerer's Diadem":
            self.element = ElementType.ANEMO
        elif self.name == "Wine-Stained Tricorne":
            self.element = ElementType.HYDRO
        else:
            assert self.name == "Witch's Scorching Hat", "Unknown name"
            self.element = ElementType.PYRO

    def value_modifier_COST(
        self,
        value: CostValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> CostValue:
        """
        When character equipped with this artifact and used skill, decrease
        the elemental cost by 1. If element not match, decrease any dice cost
        by 1.
        """
        if self.usage > 0:
            # has usage
            if not self._check_value_self_skill_or_talent(value, match):
                return value
            # can decrease cost
            if value.cost.decrease_cost(ELEMENT_TO_DIE_COLOR[self.element]):
                # decrease cost success
                if mode == "REAL":
                    self.usage -= 1
        return value


class BigElementalArtifact_4_0(SmallElementalArtifact_4_0):
    """
    Seven artifacts that decrease elemental cost and fix colors of two dice.
    """

    name: Literal[
        "Blizzard Strayer",  # cryo
        "Deepwood Memories",  # dendro
        "Archaic Petra",  # geo
        "Thundering Fury",  # electro
        "Viridescent Venerer",  # anemo
        "Heart of Depth",  # hydro
        "Crimson Witch of Flames",  # pyro
    ]
    version: Literal["4.0"] = "4.0"
    usage: int = 1
    cost: Cost = Cost(same_dice_number=2)
    element: ElementType = ElementType.NONE
    max_usage_per_round: int = 1

    def __init__(self, *argv, **kwargs):
        super(SmallElementalArtifact_4_0, self).__init__(*argv, **kwargs)
        if self.name == "Blizzard Strayer":
            self.element = ElementType.CRYO
        elif self.name == "Deepwood Memories":
            self.element = ElementType.DENDRO
        elif self.name == "Archaic Petra":
            self.element = ElementType.GEO
        elif self.name == "Thundering Fury":
            self.element = ElementType.ELECTRO
        elif self.name == "Viridescent Venerer":
            self.element = ElementType.ANEMO
        elif self.name == "Heart of Depth":
            self.element = ElementType.HYDRO
        else:
            assert self.name == "Crimson Witch of Flames", "Unknown name"
            self.element = ElementType.PYRO

    def value_modifier_INITIAL_DICE_COLOR(
        self, value: InitialDiceColorValue, match: Any, mode: Literal["REAL", "TEST"]
    ) -> InitialDiceColorValue:
        """
        If self equipped with this artifact, fix two dice colors to self
        """
        if self.position.area != ObjectPositionType.CHARACTER:
            # not in character area, do nothing
            return value
        if value.position.player_idx != self.position.player_idx:
            # not self player
            return value
        value.dice_colors += [ELEMENT_TO_DIE_COLOR[self.element]] * 2
        return value


class SmallElementalArtifact_3_3(SmallElementalArtifact_4_0):
    version: Literal["3.3"]
    cost: Cost = Cost(same_dice_number=2)


class BigElementalArtifact_3_6(BigElementalArtifact_4_0):
    version: Literal["3.6"]
    cost: Cost = Cost(any_dice_number=3)


class BigElementalArtifact_3_3(BigElementalArtifact_4_0):
    version: Literal["3.3"]
    cost: Cost = Cost(same_dice_number=3)


register_class(
    SmallElementalArtifact_4_0
    | BigElementalArtifact_4_0
    | BigElementalArtifact_3_6
    | BigElementalArtifact_3_3
    | SmallElementalArtifact_3_3
)
